VERSION = (1, 7, 5)

def get_version():
    return '{}.{}.{}'.format(VERSION[0], VERSION[1], VERSION[2])

# This try/except clause is a hack to make the get_version method work for the
# initial setup, which will fail with an ImportError because pandas hasn't yet
# been installed. Post-setup, this provides an import shortcut to Thermostat.
try:
    from .core import Thermostat
except ImportError:
    pass



from thermostat.parallel import schedule_batches

import os
import tempfile
import zipfile
from uuid import uuid4

import numpy as np
import pandas as pd
from scipy.stats import randint

import pytest

@pytest.fixture
def metadata_filename():
    columns = [
        "thermostat_id",
        "equipment_type",
        "zipcode",
        "utc_offset",
        "interval_data_filename",
    ]

    n_thermostats = 100
    thermostat_ids = [uuid4() for i in range(n_thermostats)]
    equipment_types = randint.rvs(0, 6, size=n_thermostats)
    zipcodes = [
        "70754", "70722", "70726", "70449", "70442", # "722312" 50
        "70443", "70441", "70446", "70447", "70444", # "722312"
        "70836", "70778", "70770", "70774", "70777", # "722312"
        "70433", "70437", "70436", "70435", "70438", # "722312"
        "70744", "70748", "70462", "70465", "70466", # "722312"
        "70791", "70714", "70711", "70451", "70450", # "722312"
        "70453", "70455", "70454", "70456", "70809", # "722312"
        "70806", "70807", "70805", "70769", "70761", # "722312"
        "70402", "70403", "70401", "70737", "70730", # "722312"
        "70733", "70739", "70785", "70789", "70706", # "722312"
        "45341", "45344", "45349", "45319", "45434", # "745700" 55
        "60018", "60191", "60193", "60195", "60194", # "725300" 60
        "97473", "97449", "97493", "97467", "97459", # "726917" 65
        "60421", "60544", "60404", "60408", "60481", # "725345" 70
        "36590", "36564", "36606", "36605", "36532", # "722235" 75
        "36541", "36544", "36568", "36608", "36609", # "722230" 80
        "23106", "23060", "23229", "23222", "23294", # "724029" 85
        "13674", "13601", "13606", "13605", "13682", # "726227" 90
        "12978", "12972", "12985", "12903", "12901", # "726225" 95
        "61051", # "725326" 96
        "76207", # "722589" 97
        "36362", # "722239" 98
        "57233", # "726546" 99
        "56289", # "726547" 100
    ]
    utc_offsets = [-7 for _ in range(n_thermostats)]
    interval_data_filenames = ["thermostat_{}.csv".format(i) for i in thermostat_ids]

    df = pd.DataFrame({
        "thermostat_id": thermostat_ids,
        "equipment_type": equipment_types,
        "zipcode": zipcodes,
        "utc_offset": utc_offsets,
        "interval_data_filename": interval_data_filenames,
    }, columns=columns)


    temp_dir = tempfile.mkdtemp()
    metadata_filename = os.path.join(temp_dir, "metadata.csv")
    df.to_csv(metadata_filename, index=False)

    for interval_data_filename in df.interval_data_filename:
        fname = os.path.join(temp_dir, interval_data_filename)
        with open(fname, 'w') as f :
            f.write("INTERVAL DATA FILE CONTENT")

    return metadata_filename

def test_schedule_batches_metadata_only(metadata_filename):

    batches = schedule_batches(metadata_filename, 5)

    assert len(batches) == 5
    assert sum([len(b) for b in batches]) == 100
    assert isinstance(batches[0], pd.DataFrame)

def test_schedule_batches_zip_files(metadata_filename):

    with pytest.raises(ValueError):
        schedule_batches(metadata_filename, 5, True)

    temp_dir = tempfile.mkdtemp()
    batch_zipfile_names = schedule_batches(metadata_filename, 5, True, temp_dir)

    assert len(batch_zipfile_names) == 5
    assert isinstance(batch_zipfile_names[0], str)

    with zipfile.ZipFile(batch_zipfile_names[0]) as zf:
        assert len(zf.infolist()) == 21



from thermostat.core import Thermostat

import pandas as pd
from thermostat.stations import get_closest_station_by_zipcode

from thermostat.eeweather_wrapper import get_indexed_temperatures_eeweather
from eeweather.cache import KeyValueStore
from eeweather.exceptions import ISDDataNotAvailableError
import json

import warnings
import dateutil.parser
import os
import errno
import pytz
from multiprocessing import Pool, cpu_count
from functools import partial
import logging

try:
    NUMBER_OF_CORES = len(os.sched_getaffinity(0))
except AttributeError:
    NUMBER_OF_CORES = cpu_count()
MAX_FTP_CONNECTIONS = 3
AVAILABLE_PROCESSES = min(NUMBER_OF_CORES, MAX_FTP_CONNECTIONS)


logger = logging.getLogger(__name__)


def __prime_eeweather_cache():
    """ Primes the eemeter / eeweather caches by doing a non-existent query
    This creates the cache directories sooner than if they were created
    during normal processing (which can lead to a race condition and missing
    thermostats)
    """
    sql_json = KeyValueStore()
    if sql_json.key_exists('0') is not False:
        raise Exception("eeweather cache was not properly primed. Aborting.")


def save_json_cache(index, thermostat_id, station, cache_path=None):
    """ Saves the cached results from eeweather into a JSON file.

    Parameters
    ----------
    index : pd.DatetimeIndex
        hourly index used to compute the years needed.
    thermostat_id : str
        A unique identifier for the termostat (used for the filename)
    station : str
        Station ID used to retrieve the weather data.
    cache_path : str
        Directory path to save the cached data
    """
    if cache_path is None:
        directory = os.path.join(
            os.curdir,
            "epathermostat_weather_data")
    else:
        directory = os.path.normpath(
            cache_path)

    try:
        os.mkdir(directory)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    json_cache = {}

    sqlite_json_store = KeyValueStore()
    years = index.groupby(index.year).keys()
    for year in years:
        filename = "ISD-{station}-{year}.json".format(
                station=station,
                year=year)
        json_cache[filename] = sqlite_json_store.retrieve_json(filename)

    thermostat_filename = "{thermostat_id}.json".format(thermostat_id=thermostat_id)
    thermostat_path = os.path.join(directory, thermostat_filename)
    try:
        with open(thermostat_path, 'w') as outfile:
            json.dump(json_cache, outfile)

    except Exception as e:
        warnings.warn("Unable to write JSON file: {}".format(e))


def normalize_utc_offset(utc_offset):
    """
    Normalizes the UTC offset
    Returns the UTC offset based on the string passed in.

    Parameters
    ----------
    utc_offset : str
        String representation of the UTC offset

    Returns
    -------
    datetime timdelta offset
    """
    try:
        if int(utc_offset) == 0:
            utc_offset = "+0"
        delta = dateutil.parser.parse(
            "2000-01-01T00:00:00" + str(utc_offset)).tzinfo.utcoffset(None)
        return delta

    except (ValueError, TypeError, AttributeError) as e:
        raise TypeError("Invalid UTC offset: {} ({})".format(
           utc_offset,
           e))


def from_csv(metadata_filename, verbose=False, save_cache=False, shuffle=True, cache_path=None, quiet=None):
    """
    Creates Thermostat objects from data stored in CSV files.

    Parameters
    ----------
    metadata_filename : str
        Path to a file containing the thermostat metadata.
    verbose : boolean
        Set to True to output a more detailed log of import activity.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    shuffle: boolean
        Shuffles the thermostats to give them random ordering if desired (helps with caching).
    cache_path: str
        Directory path to save the cached data

    Returns
    -------
    thermostats : iterator over thermostat.Thermostat objects
        Thermostats imported from the given CSV input files.
    """

    if quiet:
        logging.warning('quiet argument has been deprecated. Please remove this flag from your code.')

    __prime_eeweather_cache()

    metadata = pd.read_csv(
        metadata_filename,
        dtype={
            "thermostat_id": str,
            "zipcode": str,
            "utc_offset": str,
            "equipment_type": int,
            "interval_data_filename": str
        }
    )

    # Shuffle the results to help alleviate cache issues
    if shuffle:
        logging.info("Metadata randomized to prevent collisions in cache.")
        metadata = metadata.sample(frac=1).reset_index(drop=True)

    p = Pool(AVAILABLE_PROCESSES)
    multiprocess_func_partial = partial(
            multiprocess_func,
            metadata_filename=metadata_filename,
            verbose=verbose,
            save_cache=save_cache,
            cache_path=cache_path)
    result_list = p.imap(multiprocess_func_partial, metadata.iterrows())
    p.close()
    p.join()

    # Bad thermostats return None so remove those.
    results = [x for x in result_list if x is not None]

    # Check for thermostats that were not loaded and log them
    metadata_thermostat_ids = set(metadata.thermostat_id)
    loaded_thermostat_ids = set([x.thermostat_id for x in results])
    missing_thermostats = metadata_thermostat_ids.difference(loaded_thermostat_ids)
    missing_thermostats_num = len(missing_thermostats)
    if missing_thermostats_num > 0:
        logging.warning("Unable to load {} thermostat records because of "
                        "errors. Please check the logs for the following thermostats:".format(
                            missing_thermostats_num))
        for thermostat in missing_thermostats:
            logging.warning(thermostat)

    # Convert this to an iterator to maintain compatibility
    return iter(results)


def multiprocess_func(metadata, metadata_filename, verbose=False, save_cache=False, cache_path=None):
    """ This function is a partial function for multiproccessing and shares the same arguments as from_csv.
    It is not intended to be called directly."""
    i, row = metadata
    logger.info("Importing thermostat {}".format(row.thermostat_id))
    if verbose and logger.getEffectiveLevel() > logging.INFO:
        print("Importing thermostat {}".format(row.thermostat_id))

    # make sure this thermostat type is supported.
    if row.equipment_type not in [1, 2, 3, 4, 5]:
        warnings.warn(
            "Skipping import of thermostat controlling equipment"
            " of unsupported type. (id={})".format(row.thermostat_id))
        return

    interval_data_filename = os.path.join(os.path.dirname(metadata_filename), row.interval_data_filename)

    try:
        thermostat = get_single_thermostat(
                row.thermostat_id,
                row.zipcode,
                row.equipment_type,
                row.utc_offset,
                interval_data_filename,
                save_cache=save_cache,
                cache_path=cache_path,
        )
    except ValueError as e:
        # Could not locate a station for the thermostat. Warn and skip.
        warnings.warn(
            "Skipping import of thermostat (id={}) for which "
            "a sufficient source of outdoor weather data could not"
            "be located using the given ZIP code ({}). This likely "
            "due to the discrepancy between US Postal Service ZIP "
            "codes (which do not always map well to locations) and "
            "Census Bureau ZCTAs (which usually do). Please supply "
            "a zipcode which corresponds to a US Census Bureau ZCTA."
            .format(row.thermostat_id, row.zipcode))
        return

    except ISDDataNotAvailableError as e:
        warnings.warn(
            "Skipping import of thermostat(id={} because the NCDC "
            "does not have data: {}"
            .format(row.thermostat_id, e))
        return

    except Exception as e:
        warnings.warn(
            "Skipping import of thermostat(id={}) because of "
            "the following error: {}"
            .format(row.thermostat_id, e))
        return

    return thermostat


def get_single_thermostat(thermostat_id, zipcode, equipment_type,
                          utc_offset, interval_data_filename, save_cache=False, cache_path=None):
    """ Load a single thermostat directly from an interval data file.

    Parameters
    ----------
    thermostat_id : str
        A unique identifier for the thermostat.
    zipcode : str
        The zipcode of the thermostat, e.g. `"01234"`.
    equipment_type : str
        The equipment type of the thermostat.
    utc_offset : str
        A string representing the UTC offset of the interval data, e.g. `"-0700"`.
        Could also be `"Z"` (UTC), or just `"+7"` (equivalent to `"+0700"`),
        or any other timezone format recognized by the library
        method dateutil.parser.parse.
    interval_data_filename : str
        The path to the CSV in which the interval data is stored.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    cache_path: str
        Directory path to save the cached data

    Returns
    -------
    thermostat : thermostat.Thermostat
        The loaded thermostat object.
    """
    df = pd.read_csv(interval_data_filename)

    heating, cooling, aux_emerg = _get_equipment_type(equipment_type)

    # load indices
    dates = pd.to_datetime(df["date"])
    daily_index = pd.date_range(start=dates[0], periods=dates.shape[0], freq="D")
    hourly_index = pd.date_range(start=dates[0], periods=dates.shape[0] * 24, freq="H")
    hourly_index_utc = pd.date_range(start=dates[0], periods=dates.shape[0] * 24, freq="H", tz=pytz.UTC)

    # raise an error if dates are not aligned
    if not all(dates == daily_index):
        message = ("Dates provided for thermostat_id={} may contain some "
                   "which are out of order, missing, or duplicated.".format(thermostat_id))
        raise RuntimeError(message)

    # load hourly time series values
    temp_in = pd.Series(_get_hourly_block(df, "temp_in"), hourly_index)

    if heating:
        heating_setpoint = pd.Series(_get_hourly_block(df, "heating_setpoint"), hourly_index)
    else:
        heating_setpoint = None

    if cooling:
        cooling_setpoint = pd.Series(_get_hourly_block(df, "cooling_setpoint"), hourly_index)
    else:
        cooling_setpoint = None

    if aux_emerg:
        auxiliary_heat_runtime = pd.Series(_get_hourly_block(df, "auxiliary_heat_runtime"), hourly_index)
        emergency_heat_runtime = pd.Series(_get_hourly_block(df, "emergency_heat_runtime"), hourly_index)
    else:
        auxiliary_heat_runtime = None
        emergency_heat_runtime = None

    # load outdoor temperatures
    station = get_closest_station_by_zipcode(zipcode)

    if station is None:
        message = "Could not locate a valid source of outdoor temperature " \
                "data for ZIP code {}".format(zipcode)
        raise RuntimeError(message)

    utc_offset = normalize_utc_offset(utc_offset)
    temp_out = get_indexed_temperatures_eeweather(station, hourly_index_utc - utc_offset)
    temp_out.index = hourly_index

    # Export the data from the cache
    if save_cache:
        save_json_cache(hourly_index, thermostat_id, station, cache_path)

    # load daily time series values
    if cooling:
        cool_runtime = pd.Series(df["cool_runtime"].values, daily_index)
    else:
        cool_runtime = None
    if heating:
        heat_runtime = pd.Series(df["heat_runtime"].values, daily_index)
    else:
        heat_runtime = None

    # create thermostat instance
    thermostat = Thermostat(
        thermostat_id,
        equipment_type,
        zipcode,
        station,
        temp_in,
        temp_out,
        cooling_setpoint,
        heating_setpoint,
        cool_runtime,
        heat_runtime,
        auxiliary_heat_runtime,
        emergency_heat_runtime
    )
    return thermostat


def _get_hourly_block(df, prefix):
    columns = ["{}_{:02d}".format(prefix, i) for i in range(24)]
    values = df[columns].values
    return values.reshape((values.shape[0] * values.shape[1],))


def _get_equipment_type(equipment_type):
    """
    Returns
    -------
    heating : boolean
        True if the equipment type has heating equipment
    cooling : boolean
        True if the equipment type has cooling equipment
    aux_emerg : boolean
        True if the equipment type has auxiliary/emergency heat equipment
    """
    # heating, cooling, aux_emerg
    equipment_type_dict = {
        1: (True, True, True),
        2: (True, True, False),
        3: (True, True, False),
        4: (True, False, False),
        5: (False, True, False),
        }

    return(equipment_type_dict.get(equipment_type, None))



import os
import inspect

def get_data_path(f=''):
    """Return the path of a data file, these are relative to the current test
    directory. (Thanks, pydata/pandas/pandas/util/testing.py!)
    """
    # get our callers file
    _, filename, _, _, _, _ = inspect.getouterframes(inspect.currentframe())[1]
    base_dir = os.path.abspath(os.path.dirname(filename))
    return os.path.join(base_dir, f)



import logging
import json
from pkg_resources import resource_stream
from eeweather import (
        get_isd_file_metadata,
        zcta_to_lat_long,
        rank_stations,
        select_station)
from eeweather.exceptions import (
        UnrecognizedZCTAError,
        UnrecognizedUSAFIDError)

logging.getLogger(__name__)

# This is a JSON file with the zipcode to usaf_station mapping that were previously in eemeter.
zipcode_usaf_json = resource_stream('thermostat.resources', 'zipcode_usaf_station.json').read().decode()
zipcode_usaf = json.loads(zipcode_usaf_json)

# Sort order for rough_quality (returned by eeweather).
QUALITY_SORT = {'high': 0, 'medium': 1, 'low': 2}


def _rank_stations_by_distance_and_quality(lat, lon):
    """ Ranks the stations by distance and quality based on latitude / longitude

    Parameters
    ----------
    lat : float
        latitude for the search
    lon : float
        longitude for the search

    Returns
    -------
    station_ranking : Pandas.DataFrame
    """

    station_ranking = rank_stations(lat, lon)
    station_ranking['enumerated_quality'] = station_ranking['rough_quality'].map(QUALITY_SORT)
    station_ranking = station_ranking.sort_values(by=['distance_meters', 'enumerated_quality'])
    return station_ranking


def _get_closest_station_by_zcta_ranked(zcta):
    """ Selects the nth ranked station from a list of ranked stations

    Parameters
    ----------
    zcta : string
        ZIP Code Tabulation Area (ZCTA)

    Returns
    -------
    station : string
        Station that was found
    warnings : list
        List of warnings for the returned station (includes distance warnings)
    lat : float
        latitude for the search
    lon : float
        longitude for the search
    """

    zcta = zcta.zfill(5)  # Ensure that we have 5 characters, and if not left-pad it with zeroes.
    lat, lon = zcta_to_lat_long(zcta)
    finding_station = True
    rank = 0
    while finding_station:
        rank = rank + 1
        station_ranking = _rank_stations_by_distance_and_quality(lat, lon)
        station, warnings = select_station(station_ranking, rank=rank)

        # Ignore stations that begin with A
        if str(station)[0] != 'A':
            finding_station = False

    return station, warnings, lat, lon


def get_closest_station_by_zipcode(zipcode):
    """ Look up the station by ZCTA / Zip Code from eeweather
    Searches for a particular station using the ZCTA lookup for a particular zip code (from eeweather).
    The algorithm is as follows:

    1. Get a ranking of stations by distance and rough quality (quality is defined as high, medium, or low).

    2. Select the station that is the closest with the highest quality.

    3. Check to see if there are any ISD files associated with the station. If not revert to the backup method.

    4. If the station is unrecognized (:code:`UnrecognizedUSAFIDError`) then revert to the backup method.

    5. If the ZCTA is unrecognized (:code:`UnrecognizedZCTAError`) then return `None`. No station will be selected.

    6. If the station selected is not the station that would be selected via
       the zip code method then we log a debug message for what we would have previously returned

    7. If the station is over 50,000 meters from the ZCTA location then we log a warning message and revert to the backup method.

    Parameters
    ----------
    zipcode : string
        zipcode / ZCTA to look up (referred to as zipcode because this is also used for the backup method)

    Returns
    -------
    station : string or None
        Station that maps to the specified zipcode / ZCTA
    """

    station_lookup_method_by_zipcode = lookup_usaf_station_by_zipcode(zipcode)
    try:
        station, warnings, lat, lon = _get_closest_station_by_zcta_ranked(zipcode)

        isd_metadata = get_isd_file_metadata(str(station))
        if len(isd_metadata) == 0:
            logging.warning("Zipcode %s mapped to station %s, but no ISD metadata was found." % (zipcode, station))
            return station_lookup_method_by_zipcode

    except UnrecognizedUSAFIDError as e:
        logging.warning("Closest station %s is not a recognized station. Using backup-method station %s for zipcode %s instead." % (
            str(station),
            station_lookup_method_by_zipcode,
            zipcode))
        return station_lookup_method_by_zipcode

    except UnrecognizedZCTAError as e:
        logging.warning("Unrecognized ZCTA %s" % e)
        return None

    if str(station) != station_lookup_method_by_zipcode:
        logging.debug("Previously would have selected station %s instead of %s for zip code %s" % (
            station_lookup_method_by_zipcode,
            str(station),
            zipcode))

    if warnings:
        logging.warning("Station %s is %d meters over maximum %d meters (%d meters) (zip code %s is at lat/lon %f, %f)" % (
            str(station),
            int(warnings[0].data['distance_meters'] - warnings[0].data['max_distance_meters']),
            int(warnings[0].data['max_distance_meters']),
            int(warnings[0].data['distance_meters']),
            zipcode,
            lat,
            lon,
            ))
        logging.warning("Closest station %s is too far. Using backup-method station %s instead." % (
            str(station),
            station_lookup_method_by_zipcode))
        return station_lookup_method_by_zipcode

    return str(station)


def lookup_usaf_station_by_zipcode(zipcode):
    """ Backup method for determining which station matches which zipcode

    Parameters
    ----------

    zipcode : string
        zipcode to search

    Returns
    -------

    station : string or None
       Station that matches the zipcode
    """

    usaf = zipcode_usaf.get(zipcode, None)
    return usaf



from thermostat.importers import from_csv
from thermostat.importers import normalize_utc_offset
from thermostat.util.testing import get_data_path
import datetime

import pandas as pd

import pytest

from .fixtures.thermostats import (
        thermostat_type_1,
        thermostat_type_1_utc,
        thermostat_type_1_utc_bad,)

def test_import_csv(thermostat_type_1):

    def assert_is_series_with_shape(series, shape):
        assert isinstance(series, pd.Series)
        assert series.shape == shape

    assert_is_series_with_shape(thermostat_type_1.cool_runtime, (1461,))
    assert_is_series_with_shape(thermostat_type_1.heat_runtime, (1461,))

    assert_is_series_with_shape(thermostat_type_1.auxiliary_heat_runtime, (35064,))
    assert_is_series_with_shape(thermostat_type_1.emergency_heat_runtime, (35064,))

    assert_is_series_with_shape(thermostat_type_1.cooling_setpoint, (35064,))
    assert_is_series_with_shape(thermostat_type_1.heating_setpoint, (35064,))

    assert_is_series_with_shape(thermostat_type_1.temperature_in, (35064,))
    assert_is_series_with_shape(thermostat_type_1.temperature_out, (35064,))

def test_utc_offset(thermostat_type_1_utc, thermostat_type_1_utc_bad):
    assert(normalize_utc_offset("+0") == datetime.timedelta(0))
    assert(normalize_utc_offset("-0") == datetime.timedelta(0))
    assert(normalize_utc_offset("0") == datetime.timedelta(0))
    assert(normalize_utc_offset(0) == datetime.timedelta(0))
    assert(normalize_utc_offset("+6") == datetime.timedelta(0, 21600))
    assert(normalize_utc_offset("-6") == datetime.timedelta(-1, 64800))
    assert(normalize_utc_offset(-6) == datetime.timedelta(-1, 64800))

    with pytest.raises(TypeError) as excinfo:
        normalize_utc_offset(6)
    assert "Invalid UTC" in str(excinfo)

    # Load a thermostat with utc offset == 0
    assert(isinstance(thermostat_type_1_utc.cool_runtime, pd.Series))
    assert(thermostat_type_1_utc_bad is None)



import pandas as pd
import numpy as np
from numpy.testing import assert_allclose

import pytest

from .fixtures.thermostats import thermostat_type_1
from .fixtures.thermostats import core_heating_day_set_type_1_entire as core_heating_day_set_type_1
from .fixtures.thermostats import core_cooling_day_set_type_1_entire as core_cooling_day_set_type_1
from .fixtures.thermostats import core_heating_day_set_type_1_empty
from .fixtures.thermostats import core_cooling_day_set_type_1_empty
from .fixtures.thermostats import metrics_type_1_data

RTOL = 1e-3
ATOL = 1e-3


def test_get_cooling_demand_dailyavgCTD_empty(thermostat_type_1, core_cooling_day_set_type_1_empty, metrics_type_1_data):
    thermostat_type_1.get_cooling_demand(core_cooling_day_set_type_1_empty)


def test_get_cooling_demand_dailyavgHTD_empty(thermostat_type_1, core_heating_day_set_type_1_empty, metrics_type_1_data):
    thermostat_type_1.get_heating_demand(core_heating_day_set_type_1_empty)


def test_get_cooling_demand(thermostat_type_1, core_cooling_day_set_type_1, metrics_type_1_data):
    demand, tau_estimate, alpha_estimate, mse, rmse, cvrmse, mape, mae = \
            thermostat_type_1.get_cooling_demand(core_cooling_day_set_type_1)
    assert_allclose(demand.mean(), metrics_type_1_data[0]["mean_demand"], rtol=RTOL, atol=ATOL)


def test_get_heating_demand(thermostat_type_1, core_heating_day_set_type_1, metrics_type_1_data):
    demand, tau_estimate, alpha_estimate, mse, rmse, cvrmse, mape, mae = \
            thermostat_type_1.get_heating_demand(core_heating_day_set_type_1)
    assert_allclose(demand.mean(), metrics_type_1_data[1]["mean_demand"], rtol=RTOL, atol=ATOL)



from thermostat.stats import combine_output_dataframes
from thermostat.stats import compute_summary_statistics
from thermostat.stats import summary_statistics_to_csv
from .fixtures.thermostats import thermostat_emg_aux_constant_on_outlier
from thermostat.multiple import multiple_thermostat_calculate_epa_field_savings_metrics
from thermostat.exporters import COLUMNS

from scipy.stats import norm, randint
import pandas as pd
import numpy as np
import json
from datetime import datetime

import tempfile
from itertools import islice, cycle

import pytest


def get_fake_output_df(n_columns):
    columns = [
        'sw_version',

        'ct_identifier',
        'equipment_type',
        'heating_or_cooling',
        'station',
        'zipcode',
        'climate_zone',

        'start_date',
        'end_date',

        'n_days_in_inputfile_date_range',
        'n_days_both_heating_and_cooling',
        'n_days_insufficient_data',
        'n_core_cooling_days',
        'n_core_heating_days',

        'baseline_percentile_core_cooling_comfort_temperature',
        'baseline_percentile_core_heating_comfort_temperature',
        'regional_average_baseline_cooling_comfort_temperature',
        'regional_average_baseline_heating_comfort_temperature',

        'percent_savings_baseline_percentile',
        'avoided_daily_mean_core_day_runtime_baseline_percentile',
        'avoided_total_core_day_runtime_baseline_percentile',
        'baseline_daily_mean_core_day_runtime_baseline_percentile',
        'baseline_total_core_day_runtime_baseline_percentile',
        '_daily_mean_core_day_demand_baseline_baseline_percentile',
        'percent_savings_baseline_regional',
        'avoided_daily_mean_core_day_runtime_baseline_regional',
        'avoided_total_core_day_runtime_baseline_regional',
        'baseline_daily_mean_core_day_runtime_baseline_regional',
        'baseline_total_core_day_runtime_baseline_regional',
        '_daily_mean_core_day_demand_baseline_baseline_regional',
        'mean_demand',
        'alpha',
        'tau',
        'mean_sq_err',
        'root_mean_sq_err',
        'cv_root_mean_sq_err',
        'mean_abs_err',
        'mean_abs_pct_err',

        'total_core_cooling_runtime',
        'total_core_heating_runtime',
        'total_auxiliary_heating_core_day_runtime',
        'total_emergency_heating_core_day_runtime',

        'daily_mean_core_cooling_runtime',
        'daily_mean_core_heating_runtime',

        'core_cooling_days_mean_indoor_temperature',
        'core_cooling_days_mean_outdoor_temperature',
        'core_heating_days_mean_indoor_temperature',
        'core_heating_days_mean_outdoor_temperature',
        'core_mean_indoor_temperature',
        'core_mean_outdoor_temperature',

        'rhu1_aux_duty_cycle',
        'rhu1_emg_duty_cycle',
        'rhu1_compressor_duty_cycle',

        'rhu1_00F_to_05F',
        'rhu1_05F_to_10F',
        'rhu1_10F_to_15F',
        'rhu1_15F_to_20F',
        'rhu1_20F_to_25F',
        'rhu1_25F_to_30F',
        'rhu1_30F_to_35F',
        'rhu1_35F_to_40F',
        'rhu1_40F_to_45F',
        'rhu1_45F_to_50F',
        'rhu1_50F_to_55F',
        'rhu1_55F_to_60F',

        'rhu1_less10F',
        'rhu1_10F_to_20F',
        'rhu1_20F_to_30F',
        'rhu1_30F_to_40F',
        'rhu1_40F_to_50F',
        'rhu1_50F_to_60F',

        'rhu1_00F_to_05F_aux_duty_cycle',
        'rhu1_05F_to_10F_aux_duty_cycle',
        'rhu1_10F_to_15F_aux_duty_cycle',
        'rhu1_15F_to_20F_aux_duty_cycle',
        'rhu1_20F_to_25F_aux_duty_cycle',
        'rhu1_25F_to_30F_aux_duty_cycle',
        'rhu1_30F_to_35F_aux_duty_cycle',
        'rhu1_35F_to_40F_aux_duty_cycle',
        'rhu1_40F_to_45F_aux_duty_cycle',
        'rhu1_45F_to_50F_aux_duty_cycle',
        'rhu1_50F_to_55F_aux_duty_cycle',
        'rhu1_55F_to_60F_aux_duty_cycle',

        'rhu1_less10F_aux_duty_cycle',
        'rhu1_10F_to_20F_aux_duty_cycle',
        'rhu1_20F_to_30F_aux_duty_cycle',
        'rhu1_30F_to_40F_aux_duty_cycle',
        'rhu1_40F_to_50F_aux_duty_cycle',
        'rhu1_50F_to_60F_aux_duty_cycle',

        'rhu1_00F_to_05F_emg_duty_cycle',
        'rhu1_05F_to_10F_emg_duty_cycle',
        'rhu1_10F_to_15F_emg_duty_cycle',
        'rhu1_15F_to_20F_emg_duty_cycle',
        'rhu1_20F_to_25F_emg_duty_cycle',
        'rhu1_25F_to_30F_emg_duty_cycle',
        'rhu1_30F_to_35F_emg_duty_cycle',
        'rhu1_35F_to_40F_emg_duty_cycle',
        'rhu1_40F_to_45F_emg_duty_cycle',
        'rhu1_45F_to_50F_emg_duty_cycle',
        'rhu1_50F_to_55F_emg_duty_cycle',
        'rhu1_55F_to_60F_emg_duty_cycle',

        'rhu1_less10F_emg_duty_cycle',
        'rhu1_10F_to_20F_emg_duty_cycle',
        'rhu1_20F_to_30F_emg_duty_cycle',
        'rhu1_30F_to_40F_emg_duty_cycle',
        'rhu1_40F_to_50F_emg_duty_cycle',
        'rhu1_50F_to_60F_emg_duty_cycle',

        'rhu1_00F_to_05F_compressor_duty_cycle',
        'rhu1_05F_to_10F_compressor_duty_cycle',
        'rhu1_10F_to_15F_compressor_duty_cycle',
        'rhu1_15F_to_20F_compressor_duty_cycle',
        'rhu1_20F_to_25F_compressor_duty_cycle',
        'rhu1_25F_to_30F_compressor_duty_cycle',
        'rhu1_30F_to_35F_compressor_duty_cycle',
        'rhu1_35F_to_40F_compressor_duty_cycle',
        'rhu1_40F_to_45F_compressor_duty_cycle',
        'rhu1_45F_to_50F_compressor_duty_cycle',
        'rhu1_50F_to_55F_compressor_duty_cycle',
        'rhu1_55F_to_60F_compressor_duty_cycle',

        'rhu1_less10F_compressor_duty_cycle',
        'rhu1_10F_to_20F_compressor_duty_cycle',
        'rhu1_20F_to_30F_compressor_duty_cycle',
        'rhu1_30F_to_40F_compressor_duty_cycle',
        'rhu1_40F_to_50F_compressor_duty_cycle',
        'rhu1_50F_to_60F_compressor_duty_cycle',

        'rhu2_aux_duty_cycle',
        'rhu2_emg_duty_cycle',
        'rhu2_compressor_duty_cycle',

        'rhu2_00F_to_05F',
        'rhu2_05F_to_10F',
        'rhu2_10F_to_15F',
        'rhu2_15F_to_20F',
        'rhu2_20F_to_25F',
        'rhu2_25F_to_30F',
        'rhu2_30F_to_35F',
        'rhu2_35F_to_40F',
        'rhu2_40F_to_45F',
        'rhu2_45F_to_50F',
        'rhu2_50F_to_55F',
        'rhu2_55F_to_60F',

        'rhu2_less10F',
        'rhu2_10F_to_20F',
        'rhu2_20F_to_30F',
        'rhu2_30F_to_40F',
        'rhu2_40F_to_50F',
        'rhu2_50F_to_60F',

        'rhu2_00F_to_05F_aux_duty_cycle',
        'rhu2_05F_to_10F_aux_duty_cycle',
        'rhu2_10F_to_15F_aux_duty_cycle',
        'rhu2_15F_to_20F_aux_duty_cycle',
        'rhu2_20F_to_25F_aux_duty_cycle',
        'rhu2_25F_to_30F_aux_duty_cycle',
        'rhu2_30F_to_35F_aux_duty_cycle',
        'rhu2_35F_to_40F_aux_duty_cycle',
        'rhu2_40F_to_45F_aux_duty_cycle',
        'rhu2_45F_to_50F_aux_duty_cycle',
        'rhu2_50F_to_55F_aux_duty_cycle',
        'rhu2_55F_to_60F_aux_duty_cycle',

        'rhu2_less10F_aux_duty_cycle',
        'rhu2_10F_to_20F_aux_duty_cycle',
        'rhu2_20F_to_30F_aux_duty_cycle',
        'rhu2_30F_to_40F_aux_duty_cycle',
        'rhu2_40F_to_50F_aux_duty_cycle',
        'rhu2_50F_to_60F_aux_duty_cycle',

        'rhu2_00F_to_05F_emg_duty_cycle',
        'rhu2_05F_to_10F_emg_duty_cycle',
        'rhu2_10F_to_15F_emg_duty_cycle',
        'rhu2_15F_to_20F_emg_duty_cycle',
        'rhu2_20F_to_25F_emg_duty_cycle',
        'rhu2_25F_to_30F_emg_duty_cycle',
        'rhu2_30F_to_35F_emg_duty_cycle',
        'rhu2_35F_to_40F_emg_duty_cycle',
        'rhu2_40F_to_45F_emg_duty_cycle',
        'rhu2_45F_to_50F_emg_duty_cycle',
        'rhu2_50F_to_55F_emg_duty_cycle',
        'rhu2_55F_to_60F_emg_duty_cycle',

        'rhu2_less10F_emg_duty_cycle',
        'rhu2_10F_to_20F_emg_duty_cycle',
        'rhu2_20F_to_30F_emg_duty_cycle',
        'rhu2_30F_to_40F_emg_duty_cycle',
        'rhu2_40F_to_50F_emg_duty_cycle',
        'rhu2_50F_to_60F_emg_duty_cycle',

        'rhu2_00F_to_05F_compressor_duty_cycle',
        'rhu2_05F_to_10F_compressor_duty_cycle',
        'rhu2_10F_to_15F_compressor_duty_cycle',
        'rhu2_15F_to_20F_compressor_duty_cycle',
        'rhu2_20F_to_25F_compressor_duty_cycle',
        'rhu2_25F_to_30F_compressor_duty_cycle',
        'rhu2_30F_to_35F_compressor_duty_cycle',
        'rhu2_35F_to_40F_compressor_duty_cycle',
        'rhu2_40F_to_45F_compressor_duty_cycle',
        'rhu2_45F_to_50F_compressor_duty_cycle',
        'rhu2_50F_to_55F_compressor_duty_cycle',
        'rhu2_55F_to_60F_compressor_duty_cycle',

        'rhu2_less10F_compressor_duty_cycle',
        'rhu2_10F_to_20F_compressor_duty_cycle',
        'rhu2_20F_to_30F_compressor_duty_cycle',
        'rhu2_30F_to_40F_compressor_duty_cycle',
        'rhu2_40F_to_50F_compressor_duty_cycle',
        'rhu2_50F_to_60F_compressor_duty_cycle',

    ]

    string_placeholder = ["PLACEHOLDER"] * n_columns
    zero_column = [0 if randint.rvs(0, 30) > 0 else  (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in randint.rvs(0, 1, size=n_columns)]
    one_column = [1 if randint.rvs(0, 30) > 0 else  (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in randint.rvs(0, 1, size=n_columns)]
    float_column = [i if randint.rvs(0, 30) > 0 else (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in norm.rvs(size=n_columns)]
    zipcodes = ["01234", "12345", "23456", "34567", "43210", "54321", "65432", "76543"]
    zipcode_column = [i for i in islice(cycle(zipcodes), None, n_columns)]
    core_day_set_names = ["cooling_2012", "heating_2012-2013", "cooling_2013"]
    core_day_set_name_column = [i for i in islice(cycle(core_day_set_names), None, n_columns)]

    data = {
        'sw_version': string_placeholder,

        'ct_identifier': string_placeholder,
        'equipment_type': string_placeholder,
        'heating_or_cooling': core_day_set_name_column,
        'station': string_placeholder,
        'zipcode': zipcode_column,
        'climate_zone': string_placeholder,

        'start_date': datetime(2011, 1, 1),
        'end_date': datetime(2012, 1, 1),
        'n_days_both_heating_and_cooling': one_column,
        'n_days_in_inputfile_date_range': one_column,
        'n_days_insufficient_data': zero_column,
        'n_core_heating_days': one_column,

        'baseline_percentile_core_cooling_comfort_temperature': float_column,
        'baseline_percentile_core_heating_comfort_temperature': float_column,
        'regional_average_baseline_cooling_comfort_temperature': float_column,
        'regional_average_baseline_heating_comfort_temperature': float_column,

        'percent_savings_baseline_percentile': float_column,
        'avoided_daily_mean_core_day_runtime_baseline_percentile': float_column,
        'avoided_total_core_day_runtime_baseline_percentile': float_column,
        'baseline_daily_mean_core_day_runtime_baseline_percentile': float_column,
        'baseline_total_core_day_runtime_baseline_percentile': float_column,
        '_daily_mean_core_day_demand_baseline_baseline_percentile': float_column,
        'percent_savings_baseline_regional': float_column,
        'avoided_daily_mean_core_day_runtime_baseline_regional': float_column,
        'avoided_total_core_day_runtime_baseline_regional': float_column,
        'baseline_daily_mean_core_day_runtime_baseline_regional': float_column,
        'baseline_total_core_day_runtime_baseline_regional': float_column,
        '_daily_mean_core_day_demand_baseline_baseline_regional': float_column,
        'mean_demand': float_column,
        'alpha': float_column,
        'tau': float_column,
        'mean_sq_err': float_column,
        'root_mean_sq_err': float_column,
        'cv_root_mean_sq_err': float_column,
        'mean_abs_err': float_column,
        'mean_abs_pct_err': float_column,

        'total_core_cooling_runtime': float_column,
        'total_core_heating_runtime': float_column,
        'total_auxiliary_heating_core_day_runtime': float_column,
        'total_emergency_heating_core_day_runtime': float_column,

        'daily_mean_core_cooling_runtime': float_column,
        'daily_mean_core_heating_runtime': float_column,

        'core_cooling_days_mean_indoor_temperature': float_column,
        'core_cooling_days_mean_outdoor_temperature': float_column,
        'core_heating_days_mean_indoor_temperature': float_column,
        'core_heating_days_mean_outdoor_temperature': float_column,
        'core_mean_indoor_temperature': float_column,
        'core_mean_outdoor_temperature': float_column,

        'rhu1_aux_duty_cycle': float_column,
        'rhu1_emg_duty_cycle': float_column,
        'rhu1_compressor_duty_cycle': float_column,

        'rhu1_00F_to_05F': float_column,
        'rhu1_05F_to_10F': float_column,
        'rhu1_10F_to_15F': float_column,
        'rhu1_15F_to_20F': float_column,
        'rhu1_20F_to_25F': float_column,
        'rhu1_25F_to_30F': float_column,
        'rhu1_30F_to_35F': float_column,
        'rhu1_35F_to_40F': float_column,
        'rhu1_40F_to_45F': float_column,
        'rhu1_45F_to_50F': float_column,
        'rhu1_50F_to_55F': float_column,
        'rhu1_55F_to_60F': float_column,

        'rhu1_less10F': float_column,
        'rhu1_10F_to_20F': float_column,
        'rhu1_20F_to_30F': float_column,
        'rhu1_30F_to_40F': float_column,
        'rhu1_40F_to_50F': float_column,
        'rhu1_50F_to_60F': float_column,

        'rhu1_00F_to_05F_aux_duty_cycle': float_column,
        'rhu1_05F_to_10F_aux_duty_cycle': float_column,
        'rhu1_10F_to_15F_aux_duty_cycle': float_column,
        'rhu1_15F_to_20F_aux_duty_cycle': float_column,
        'rhu1_20F_to_25F_aux_duty_cycle': float_column,
        'rhu1_25F_to_30F_aux_duty_cycle': float_column,
        'rhu1_30F_to_35F_aux_duty_cycle': float_column,
        'rhu1_35F_to_40F_aux_duty_cycle': float_column,
        'rhu1_40F_to_45F_aux_duty_cycle': float_column,
        'rhu1_45F_to_50F_aux_duty_cycle': float_column,
        'rhu1_50F_to_55F_aux_duty_cycle': float_column,
        'rhu1_55F_to_60F_aux_duty_cycle': float_column,

        'rhu1_less10F_aux_duty_cycle': float_column,
        'rhu1_10F_to_20F_aux_duty_cycle': float_column,
        'rhu1_20F_to_30F_aux_duty_cycle': float_column,
        'rhu1_30F_to_40F_aux_duty_cycle': float_column,
        'rhu1_40F_to_50F_aux_duty_cycle': float_column,
        'rhu1_50F_to_60F_aux_duty_cycle': float_column,

        'rhu1_00F_to_05F_emg_duty_cycle': float_column,
        'rhu1_05F_to_10F_emg_duty_cycle': float_column,
        'rhu1_10F_to_15F_emg_duty_cycle': float_column,
        'rhu1_15F_to_20F_emg_duty_cycle': float_column,
        'rhu1_20F_to_25F_emg_duty_cycle': float_column,
        'rhu1_25F_to_30F_emg_duty_cycle': float_column,
        'rhu1_30F_to_35F_emg_duty_cycle': float_column,
        'rhu1_35F_to_40F_emg_duty_cycle': float_column,
        'rhu1_40F_to_45F_emg_duty_cycle': float_column,
        'rhu1_45F_to_50F_emg_duty_cycle': float_column,
        'rhu1_50F_to_55F_emg_duty_cycle': float_column,
        'rhu1_55F_to_60F_emg_duty_cycle': float_column,

        'rhu1_less10F_emg_duty_cycle': float_column,
        'rhu1_10F_to_20F_emg_duty_cycle': float_column,
        'rhu1_20F_to_30F_emg_duty_cycle': float_column,
        'rhu1_30F_to_40F_emg_duty_cycle': float_column,
        'rhu1_40F_to_50F_emg_duty_cycle': float_column,
        'rhu1_50F_to_60F_emg_duty_cycle': float_column,

        'rhu1_00F_to_05F_compressor_duty_cycle': float_column,
        'rhu1_05F_to_10F_compressor_duty_cycle': float_column,
        'rhu1_10F_to_15F_compressor_duty_cycle': float_column,
        'rhu1_15F_to_20F_compressor_duty_cycle': float_column,
        'rhu1_20F_to_25F_compressor_duty_cycle': float_column,
        'rhu1_25F_to_30F_compressor_duty_cycle': float_column,
        'rhu1_30F_to_35F_compressor_duty_cycle': float_column,
        'rhu1_35F_to_40F_compressor_duty_cycle': float_column,
        'rhu1_40F_to_45F_compressor_duty_cycle': float_column,
        'rhu1_45F_to_50F_compressor_duty_cycle': float_column,
        'rhu1_50F_to_55F_compressor_duty_cycle': float_column,
        'rhu1_55F_to_60F_compressor_duty_cycle': float_column,

        'rhu1_less10F_compressor_duty_cycle': float_column,
        'rhu1_10F_to_20F_compressor_duty_cycle': float_column,
        'rhu1_20F_to_30F_compressor_duty_cycle': float_column,
        'rhu1_30F_to_40F_compressor_duty_cycle': float_column,
        'rhu1_40F_to_50F_compressor_duty_cycle': float_column,
        'rhu1_50F_to_60F_compressor_duty_cycle': float_column,

        'rhu2_aux_duty_cycle': float_column,
        'rhu2_emg_duty_cycle': float_column,
        'rhu2_compressor_duty_cycle': float_column,

        'rhu2_00F_to_05F': float_column,
        'rhu2_05F_to_10F': float_column,
        'rhu2_10F_to_15F': float_column,
        'rhu2_15F_to_20F': float_column,
        'rhu2_20F_to_25F': float_column,
        'rhu2_25F_to_30F': float_column,
        'rhu2_30F_to_35F': float_column,
        'rhu2_35F_to_40F': float_column,
        'rhu2_40F_to_45F': float_column,
        'rhu2_45F_to_50F': float_column,
        'rhu2_50F_to_55F': float_column,
        'rhu2_55F_to_60F': float_column,

        'rhu2_less10F': float_column,
        'rhu2_10F_to_20F': float_column,
        'rhu2_20F_to_30F': float_column,
        'rhu2_30F_to_40F': float_column,
        'rhu2_40F_to_50F': float_column,
        'rhu2_50F_to_60F': float_column,

        'rhu2_00F_to_05F_aux_duty_cycle': float_column,
        'rhu2_05F_to_10F_aux_duty_cycle': float_column,
        'rhu2_10F_to_15F_aux_duty_cycle': float_column,
        'rhu2_15F_to_20F_aux_duty_cycle': float_column,
        'rhu2_20F_to_25F_aux_duty_cycle': float_column,
        'rhu2_25F_to_30F_aux_duty_cycle': float_column,
        'rhu2_30F_to_35F_aux_duty_cycle': float_column,
        'rhu2_35F_to_40F_aux_duty_cycle': float_column,
        'rhu2_40F_to_45F_aux_duty_cycle': float_column,
        'rhu2_45F_to_50F_aux_duty_cycle': float_column,
        'rhu2_50F_to_55F_aux_duty_cycle': float_column,
        'rhu2_55F_to_60F_aux_duty_cycle': float_column,

        'rhu2_less10F_aux_duty_cycle': float_column,
        'rhu2_10F_to_20F_aux_duty_cycle': float_column,
        'rhu2_20F_to_30F_aux_duty_cycle': float_column,
        'rhu2_30F_to_40F_aux_duty_cycle': float_column,
        'rhu2_40F_to_50F_aux_duty_cycle': float_column,
        'rhu2_50F_to_60F_aux_duty_cycle': float_column,

        'rhu2_00F_to_05F_emg_duty_cycle': float_column,
        'rhu2_05F_to_10F_emg_duty_cycle': float_column,
        'rhu2_10F_to_15F_emg_duty_cycle': float_column,
        'rhu2_15F_to_20F_emg_duty_cycle': float_column,
        'rhu2_20F_to_25F_emg_duty_cycle': float_column,
        'rhu2_25F_to_30F_emg_duty_cycle': float_column,
        'rhu2_30F_to_35F_emg_duty_cycle': float_column,
        'rhu2_35F_to_40F_emg_duty_cycle': float_column,
        'rhu2_40F_to_45F_emg_duty_cycle': float_column,
        'rhu2_45F_to_50F_emg_duty_cycle': float_column,
        'rhu2_50F_to_55F_emg_duty_cycle': float_column,
        'rhu2_55F_to_60F_emg_duty_cycle': float_column,

        'rhu2_less10F_emg_duty_cycle': float_column,
        'rhu2_10F_to_20F_emg_duty_cycle': float_column,
        'rhu2_20F_to_30F_emg_duty_cycle': float_column,
        'rhu2_30F_to_40F_emg_duty_cycle': float_column,
        'rhu2_40F_to_50F_emg_duty_cycle': float_column,
        'rhu2_50F_to_60F_emg_duty_cycle': float_column,

        'rhu2_00F_to_05F_compressor_duty_cycle': float_column,
        'rhu2_05F_to_10F_compressor_duty_cycle': float_column,
        'rhu2_10F_to_15F_compressor_duty_cycle': float_column,
        'rhu2_15F_to_20F_compressor_duty_cycle': float_column,
        'rhu2_20F_to_25F_compressor_duty_cycle': float_column,
        'rhu2_25F_to_30F_compressor_duty_cycle': float_column,
        'rhu2_30F_to_35F_compressor_duty_cycle': float_column,
        'rhu2_35F_to_40F_compressor_duty_cycle': float_column,
        'rhu2_40F_to_45F_compressor_duty_cycle': float_column,
        'rhu2_45F_to_50F_compressor_duty_cycle': float_column,
        'rhu2_50F_to_55F_compressor_duty_cycle': float_column,
        'rhu2_55F_to_60F_compressor_duty_cycle': float_column,

        'rhu2_less10F_compressor_duty_cycle': float_column,
        'rhu2_10F_to_20F_compressor_duty_cycle': float_column,
        'rhu2_20F_to_30F_compressor_duty_cycle': float_column,
        'rhu2_30F_to_40F_compressor_duty_cycle': float_column,
        'rhu2_40F_to_50F_compressor_duty_cycle': float_column,
        'rhu2_50F_to_60F_compressor_duty_cycle': float_column,
    }
    df = pd.DataFrame(data, columns=columns)
    return df


@pytest.fixture
def dataframes():
    df1 = get_fake_output_df(10)
    df2 = get_fake_output_df(10)
    dfs = [df1, df2]
    return dfs


@pytest.fixture
def combined_dataframe():
    df = get_fake_output_df(100)
    return df


def test_combine_output_dataframes(dataframes):
    combined = combine_output_dataframes(dataframes)
    assert combined.shape == (20, 200)


def test_compute_summary_statistics(combined_dataframe):
    summary_statistics = compute_summary_statistics(combined_dataframe)
    assert [len(s) for s in summary_statistics] == [
            49, 49, 49, 49,
            9105, 901, 9105, 901,
            ]

    def test_compute_summary_statistics_advanced(combined_dataframe):
        summary_statistics = compute_summary_statistics(combined_dataframe,
                advanced_filtering=True)
        assert [len(s) for s in summary_statistics] == [
                49, 49, 49, 49, 49, 49, 49, 49,
                9105, 901, 9105, 901, 9105, 901, 9105, 901,
                ]

        def test_summary_statistics_to_csv(combined_dataframe):
            summary_statistics = compute_summary_statistics(combined_dataframe)

    _, fname = tempfile.mkstemp()
    product_id = "FAKE"
    stats_df = summary_statistics_to_csv(summary_statistics, fname, product_id)
    assert isinstance(stats_df, pd.DataFrame)

    stats_df_reread = pd.read_csv(fname)
    assert stats_df_reread.shape == (9241, 9)


def test_iqr_filteringa(thermostat_emg_aux_constant_on_outlier):

    thermostats_iqflt = list(thermostat_emg_aux_constant_on_outlier)
    # Run the metrics / statistics with the outlier thermostat in place
    iqflt_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(thermostats_iqflt)
    iqflt_output_dataframe = pd.DataFrame(iqflt_metrics, columns=COLUMNS)
    iqflt_summary_statistics = compute_summary_statistics(iqflt_output_dataframe)

    # Remove the outlier thermostat
    thermostats_noiq = []
    for thermostat in list(thermostats_iqflt):
        if thermostat.thermostat_id != 'thermostat_single_emg_aux_constant_on_outlier':
            thermostats_noiq.append(thermostat)

    if len(thermostats_noiq) == 5:
        raise ValueError("Try again")

    # Re-run the metrics / statistics with the outlier thermostat removed
    noiq_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(thermostats_noiq)
    noiq_output_dataframe = pd.DataFrame(noiq_metrics, columns=COLUMNS)
    noiq_summary_statistics = compute_summary_statistics(noiq_output_dataframe)

    # Verify that the IQFLT removed the outliers by comparing this with the
    # metrics with the outlier thermostat already removed.
    for column in range(0, len(iqflt_summary_statistics)):
        fields_iqflt = [x for x in iqflt_summary_statistics[column] if '_IQFLT' in x]
        for field_iqflt in fields_iqflt:
            field_noiq = field_iqflt.replace('IQFLT', 'NOIQ')
            left_side = iqflt_summary_statistics[column][field_iqflt]
            right_side = noiq_summary_statistics[column][field_noiq]

            if np.isnan(left_side) or np.isnan(right_side):
                assert(np.isnan(left_side) and np.isnan(right_side))
            else:
                assert(left_side == right_side)



import pandas as pd
from pkg_resources import resource_stream
from collections import namedtuple


def _load_mapping(filename_or_buffer):
    df = pd.read_csv(
        filename_or_buffer,
        usecols=["zipcode", "group"],
        dtype={"zipcode": str, "group": str},
    ).set_index('zipcode').drop('zipcode', errors='ignore')
    df = df.where((pd.notnull(df)), None)

    return dict(df.to_records('index'))


def retrieve_climate_zone(climate_zone_mapping, zipcode):
    """ Loads the Climate Zone to Zipcode database
    and returns the climate zone and baseline regional comfort temperatures.

    Parameters
    ----------

    climate_zone_mapping : filename, default: None

        A mapping from climate zone to zipcode. If None is provided, uses
        default zipcode to climate zone mapping provided in tutorial.

        :download:`default mapping <./resources/Building America Climate Zone to Zipcode Database_Rev2_2016.09.08.csv>`

    Returns
    -------

    climate_zone_nt : named tuple
       Named Tuple consisting of the Climate Zone, baseline_regional_cooling_comfort_temperature, and baseline_regional_heating_comfort_temperature
    """

    ClimateZone = namedtuple('ClimateZone', ['climate_zone', 'baseline_regional_cooling_comfort_temperature', 'baseline_regional_heating_comfort_temperature'])
    if climate_zone_mapping is None:
        with resource_stream(
                'thermostat.resources',
                'Building America Climate Zone to Zipcode Database_Rev2_2016.09.08.csv') as f:
            mapping = _load_mapping(f)
    else:
        try:
            mapping = _load_mapping(climate_zone_mapping)
        except Exception as e:
            raise ValueError("Could not load climate zone mapping: %s" % e)

    with resource_stream('thermostat.resources', 'regional_baselines.csv') as f:
        df = pd.read_csv(
            f, usecols=[
                'EIA Climate Zone',
                'Baseline heating temp (F)',
                'Baseline cooling temp (F)'
            ])
        df = df.where((pd.notnull(df)), None)
        df = df.set_index('EIA Climate Zone')
        cooling_regional_baseline_temps = {k: v for k, v in df['Baseline cooling temp (F)'].iteritems()}
        heating_regional_baseline_temps = {k: v for k, v in df['Baseline heating temp (F)'].iteritems()}

    climate_zone = mapping.get(zipcode)
    baseline_regional_cooling_comfort_temperature = cooling_regional_baseline_temps.get(climate_zone, None)
    baseline_regional_heating_comfort_temperature = heating_regional_baseline_temps.get(climate_zone, None)

    climate_zone_nt = ClimateZone(climate_zone, baseline_regional_cooling_comfort_temperature, baseline_regional_heating_comfort_temperature)

    return climate_zone_nt



from thermostat.importers import from_csv
from thermostat.util.testing import get_data_path
from thermostat.regression import runtime_regression

import pandas as pd
import numpy as np
from numpy.testing import assert_allclose

import pytest

RTOL = 1e-3
ATOL = 1e-3

@pytest.fixture(params=[
    (pd.Series([1, 2, 3, 4, 5, 6]), pd.Series([2, 4, 6, 8, 10, 12]), .5, 0, 0),
    (pd.Series([2, 4, 10, 8, 6]), pd.Series([1, 2, 5, 4, 3]), 2, 0, 0),
    (pd.Series([4, 5, 6, 7]), pd.Series([1, 2, 3, 4]), 1, 3, 0),
    (pd.Series([4, 5, 6, np.nan]), pd.Series([1, 2, 3, 4]), 1, 3, 0),
    (pd.Series([4]), pd.Series([1]), np.nan, np.nan, np.nan),
    (pd.Series([]), pd.Series([]), np.nan, np.nan, np.nan),
    ])
def regression_fixture(request):
    return request.param

def test_runtime_regression(regression_fixture):
    daily_runtime, daily_demand, slope, intercept, mse = regression_fixture
    slope_, intercept_, mse_, rmse_, cvrmse_, mape_, mae_  = runtime_regression(daily_runtime, daily_demand, "cooling")

    assert_allclose(slope_, slope, rtol=RTOL, atol=ATOL)
    assert_allclose(intercept_, intercept, rtol=RTOL, atol=ATOL)
    assert_allclose(mse_, mse, rtol=RTOL, atol=ATOL)



import pandas as pd

COLUMNS = [
    'sw_version',

    'ct_identifier',
    'equipment_type',
    'heating_or_cooling',
    'zipcode',
    'station',
    'climate_zone',

    'start_date',
    'end_date',
    'n_days_in_inputfile_date_range',
    'n_days_both_heating_and_cooling',
    'n_days_insufficient_data',
    'n_core_cooling_days',
    'n_core_heating_days',

    'baseline_percentile_core_cooling_comfort_temperature',
    'baseline_percentile_core_heating_comfort_temperature',
    'regional_average_baseline_cooling_comfort_temperature',
    'regional_average_baseline_heating_comfort_temperature',

    'percent_savings_baseline_percentile',
    'avoided_daily_mean_core_day_runtime_baseline_percentile',
    'avoided_total_core_day_runtime_baseline_percentile',
    'baseline_daily_mean_core_day_runtime_baseline_percentile',
    'baseline_total_core_day_runtime_baseline_percentile',
    '_daily_mean_core_day_demand_baseline_baseline_percentile',
    'percent_savings_baseline_regional',
    'avoided_daily_mean_core_day_runtime_baseline_regional',
    'avoided_total_core_day_runtime_baseline_regional',
    'baseline_daily_mean_core_day_runtime_baseline_regional',
    'baseline_total_core_day_runtime_baseline_regional',
    '_daily_mean_core_day_demand_baseline_baseline_regional',
    'mean_demand',
    'alpha',
    'tau',
    'mean_sq_err',
    'root_mean_sq_err',
    'cv_root_mean_sq_err',
    'mean_abs_err',
    'mean_abs_pct_err',

    'total_core_cooling_runtime',
    'total_core_heating_runtime',
    'total_auxiliary_heating_core_day_runtime',
    'total_emergency_heating_core_day_runtime',

    'daily_mean_core_cooling_runtime',
    'daily_mean_core_heating_runtime',

    'core_cooling_days_mean_indoor_temperature',
    'core_cooling_days_mean_outdoor_temperature',
    'core_heating_days_mean_indoor_temperature',
    'core_heating_days_mean_outdoor_temperature',
    'core_mean_indoor_temperature',
    'core_mean_outdoor_temperature',

    'rhu1_aux_duty_cycle',
    'rhu1_emg_duty_cycle',
    'rhu1_compressor_duty_cycle',

    'rhu1_00F_to_05F',
    'rhu1_05F_to_10F',
    'rhu1_10F_to_15F',
    'rhu1_15F_to_20F',
    'rhu1_20F_to_25F',
    'rhu1_25F_to_30F',
    'rhu1_30F_to_35F',
    'rhu1_35F_to_40F',
    'rhu1_40F_to_45F',
    'rhu1_45F_to_50F',
    'rhu1_50F_to_55F',
    'rhu1_55F_to_60F',

    'rhu1_05F_to_10F_aux_duty_cycle',
    'rhu1_10F_to_15F_aux_duty_cycle',
    'rhu1_15F_to_20F_aux_duty_cycle',
    'rhu1_20F_to_25F_aux_duty_cycle',
    'rhu1_25F_to_30F_aux_duty_cycle',
    'rhu1_30F_to_35F_aux_duty_cycle',
    'rhu1_35F_to_40F_aux_duty_cycle',
    'rhu1_40F_to_45F_aux_duty_cycle',
    'rhu1_45F_to_50F_aux_duty_cycle',
    'rhu1_50F_to_55F_aux_duty_cycle',
    'rhu1_55F_to_60F_aux_duty_cycle',

    'rhu1_00F_to_05F_emg_duty_cycle',
    'rhu1_05F_to_10F_emg_duty_cycle',
    'rhu1_10F_to_15F_emg_duty_cycle',
    'rhu1_15F_to_20F_emg_duty_cycle',
    'rhu1_20F_to_25F_emg_duty_cycle',
    'rhu1_25F_to_30F_emg_duty_cycle',
    'rhu1_30F_to_35F_emg_duty_cycle',
    'rhu1_35F_to_40F_emg_duty_cycle',
    'rhu1_40F_to_45F_emg_duty_cycle',
    'rhu1_45F_to_50F_emg_duty_cycle',
    'rhu1_50F_to_55F_emg_duty_cycle',
    'rhu1_55F_to_60F_emg_duty_cycle',

    'rhu1_00F_to_05F_compressor_duty_cycle',
    'rhu1_05F_to_10F_compressor_duty_cycle',
    'rhu1_10F_to_15F_compressor_duty_cycle',
    'rhu1_15F_to_20F_compressor_duty_cycle',
    'rhu1_20F_to_25F_compressor_duty_cycle',
    'rhu1_25F_to_30F_compressor_duty_cycle',
    'rhu1_30F_to_35F_compressor_duty_cycle',
    'rhu1_35F_to_40F_compressor_duty_cycle',
    'rhu1_40F_to_45F_compressor_duty_cycle',
    'rhu1_45F_to_50F_compressor_duty_cycle',
    'rhu1_50F_to_55F_compressor_duty_cycle',
    'rhu1_55F_to_60F_compressor_duty_cycle',

    'rhu1_less10F',
    'rhu1_10F_to_20F',
    'rhu1_20F_to_30F',
    'rhu1_30F_to_40F',
    'rhu1_40F_to_50F',
    'rhu1_50F_to_60F',
    'rhu1_00F_to_05F_aux_duty_cycle',

    'rhu1_less10F_aux_duty_cycle',
    'rhu1_10F_to_20F_aux_duty_cycle',
    'rhu1_20F_to_30F_aux_duty_cycle',
    'rhu1_30F_to_40F_aux_duty_cycle',
    'rhu1_40F_to_50F_aux_duty_cycle',
    'rhu1_50F_to_60F_aux_duty_cycle',

    'rhu1_less10F_emg_duty_cycle',
    'rhu1_10F_to_20F_emg_duty_cycle',
    'rhu1_20F_to_30F_emg_duty_cycle',
    'rhu1_30F_to_40F_emg_duty_cycle',
    'rhu1_40F_to_50F_emg_duty_cycle',
    'rhu1_50F_to_60F_emg_duty_cycle',

    'rhu1_less10F_compressor_duty_cycle',
    'rhu1_10F_to_20F_compressor_duty_cycle',
    'rhu1_20F_to_30F_compressor_duty_cycle',
    'rhu1_30F_to_40F_compressor_duty_cycle',
    'rhu1_40F_to_50F_compressor_duty_cycle',
    'rhu1_50F_to_60F_compressor_duty_cycle',

    'rhu2_00F_to_05F',
    'rhu2_05F_to_10F',
    'rhu2_10F_to_15F',
    'rhu2_15F_to_20F',
    'rhu2_20F_to_25F',
    'rhu2_25F_to_30F',
    'rhu2_30F_to_35F',
    'rhu2_35F_to_40F',
    'rhu2_40F_to_45F',
    'rhu2_45F_to_50F',
    'rhu2_50F_to_55F',
    'rhu2_55F_to_60F',

    'rhu2_00F_to_05F_aux_duty_cycle',
    'rhu2_05F_to_10F_aux_duty_cycle',
    'rhu2_10F_to_15F_aux_duty_cycle',
    'rhu2_15F_to_20F_aux_duty_cycle',
    'rhu2_20F_to_25F_aux_duty_cycle',
    'rhu2_25F_to_30F_aux_duty_cycle',
    'rhu2_30F_to_35F_aux_duty_cycle',
    'rhu2_35F_to_40F_aux_duty_cycle',
    'rhu2_40F_to_45F_aux_duty_cycle',
    'rhu2_45F_to_50F_aux_duty_cycle',
    'rhu2_50F_to_55F_aux_duty_cycle',
    'rhu2_55F_to_60F_aux_duty_cycle',

    'rhu2_00F_to_05F_emg_duty_cycle',
    'rhu2_05F_to_10F_emg_duty_cycle',
    'rhu2_10F_to_15F_emg_duty_cycle',
    'rhu2_15F_to_20F_emg_duty_cycle',
    'rhu2_20F_to_25F_emg_duty_cycle',
    'rhu2_25F_to_30F_emg_duty_cycle',
    'rhu2_30F_to_35F_emg_duty_cycle',
    'rhu2_35F_to_40F_emg_duty_cycle',
    'rhu2_40F_to_45F_emg_duty_cycle',
    'rhu2_45F_to_50F_emg_duty_cycle',
    'rhu2_50F_to_55F_emg_duty_cycle',
    'rhu2_55F_to_60F_emg_duty_cycle',

    'rhu2_00F_to_05F_compressor_duty_cycle',
    'rhu2_05F_to_10F_compressor_duty_cycle',
    'rhu2_10F_to_15F_compressor_duty_cycle',
    'rhu2_15F_to_20F_compressor_duty_cycle',
    'rhu2_20F_to_25F_compressor_duty_cycle',
    'rhu2_25F_to_30F_compressor_duty_cycle',
    'rhu2_30F_to_35F_compressor_duty_cycle',
    'rhu2_35F_to_40F_compressor_duty_cycle',
    'rhu2_40F_to_45F_compressor_duty_cycle',
    'rhu2_45F_to_50F_compressor_duty_cycle',
    'rhu2_50F_to_55F_compressor_duty_cycle',
    'rhu2_55F_to_60F_compressor_duty_cycle',

    'rhu2_less10F',
    'rhu2_10F_to_20F',
    'rhu2_20F_to_30F',
    'rhu2_30F_to_40F',
    'rhu2_40F_to_50F',
    'rhu2_50F_to_60F',

    'rhu2_less10F_aux_duty_cycle',
    'rhu2_10F_to_20F_aux_duty_cycle',
    'rhu2_20F_to_30F_aux_duty_cycle',
    'rhu2_30F_to_40F_aux_duty_cycle',
    'rhu2_40F_to_50F_aux_duty_cycle',
    'rhu2_50F_to_60F_aux_duty_cycle',

    'rhu2_less10F_emg_duty_cycle',
    'rhu2_10F_to_20F_emg_duty_cycle',
    'rhu2_20F_to_30F_emg_duty_cycle',
    'rhu2_30F_to_40F_emg_duty_cycle',
    'rhu2_40F_to_50F_emg_duty_cycle',
    'rhu2_50F_to_60F_emg_duty_cycle',

    'rhu2_less10F_compressor_duty_cycle',
    'rhu2_10F_to_20F_compressor_duty_cycle',
    'rhu2_20F_to_30F_compressor_duty_cycle',
    'rhu2_30F_to_40F_compressor_duty_cycle',
    'rhu2_40F_to_50F_compressor_duty_cycle',
    'rhu2_50F_to_60F_compressor_duty_cycle',
    ]


def metrics_to_csv(metrics, filepath):
    """ Writes metrics outputs to the file specified.

    Parameters
    ----------
    metrics : list of dict
        list of outputs from the function
        `thermostat.calculate_epa_draft_rccs_field_savings_metrics()`
    filepath : str
        filepath specification for location of output CSV file.

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing data output to CSV.
    """

    output_dataframe = pd.DataFrame(metrics, columns=COLUMNS)
    output_dataframe.to_csv(filepath, index=False, columns=COLUMNS)
    return output_dataframe



import pandas as pd
import numpy as np

from numpy.testing import assert_allclose

import pytest

from .fixtures.thermostats import thermostat_type_1
from .fixtures.thermostats import core_heating_day_set_type_1_entire
from .fixtures.thermostats import core_cooling_day_set_type_1_entire
from .fixtures.thermostats import metrics_type_1_data
from .fixtures.thermostats import thermostat_template

from thermostat.core import CoreDaySet
from datetime import datetime

RTOL = 1e-3
ATOL = 1e-3


@pytest.mark.slow
def test_get_core_cooling_day_baseline_setpoint(thermostat_type_1, core_cooling_day_set_type_1_entire, metrics_type_1_data):

    baseline = thermostat_type_1.get_core_cooling_day_baseline_setpoint(core_cooling_day_set_type_1_entire)
    assert_allclose(baseline, metrics_type_1_data[0]["baseline_percentile_core_cooling_comfort_temperature"], rtol=RTOL, atol=ATOL)

@pytest.mark.slow
def test_get_core_heating_day_baseline_setpoint(thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data):

    baseline = thermostat_type_1.get_core_heating_day_baseline_setpoint(core_heating_day_set_type_1_entire)
    assert_allclose(baseline, metrics_type_1_data[1]["baseline_percentile_core_heating_comfort_temperature"], rtol=RTOL, atol=ATOL)

def test_get_core_heating_day_baseline_setpoint_null_temperature_in(thermostat_template):

    index = pd.date_range(start=datetime(2011, 1, 1), periods=3, freq='D')
    thermostat_template.temperature_in = pd.Series([0, 1, 2], index=index)
    thermostat_template.equipment_type = 1

    season_selector = pd.Series([True, True, True], index=index)
    core_day_set = CoreDaySet("FAKE", season_selector, season_selector, None, None)

    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='heating_setpoint')
    assert pd.isnull(sp)

    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 1.8)

    thermostat_template.temperature_in = pd.Series([np.nan, 1, 2], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 1.9)

    thermostat_template.temperature_in = pd.Series([0, 1, np.nan], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 0.9)

    thermostat_template.temperature_in = pd.Series([np.nan, np.nan, np.nan], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert pd.isnull(sp)

def test_get_core_cooling_day_baseline_setpoint_null_temperature_in(thermostat_template):

    index = pd.date_range(start=datetime(2011, 1, 1), periods=3, freq='D')
    thermostat_template.temperature_in = pd.Series([0, 1, 2], index=index)
    thermostat_template.equipment_type = 1

    season_selector = pd.Series([True, True, True], index=index)
    core_day_set = CoreDaySet("FAKE", season_selector, season_selector, None, None)

    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='cooling_setpoint')
    assert pd.isnull(sp)

    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 0.2)

    thermostat_template.temperature_in = pd.Series([np.nan, 1, 2], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 1.1)

    thermostat_template.temperature_in = pd.Series([0, 1, np.nan], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert_allclose(sp, 0.1)

    thermostat_template.temperature_in = pd.Series([np.nan, np.nan, np.nan], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert pd.isnull(sp)

def test_get_core_heating_day_baseline_setpoint_null_heating_setpoint(thermostat_template):

    index = pd.date_range(start=datetime(2011, 1, 1), periods=3, freq='D')
    thermostat_template.heating_setpoint = pd.Series([0, 1, 2], index=index)
    thermostat_template.equipment_type = 1

    season_selector = pd.Series([True, True, True], index=index)
    core_day_set = CoreDaySet("FAKE", season_selector, season_selector, None, None)

    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert pd.isnull(sp)

    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='heating_setpoint')
    assert_allclose(sp, 1.8)

    thermostat_template.heating_setpoint = pd.Series([np.nan, 1, 2], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='heating_setpoint')
    assert_allclose(sp, 1.9)

    thermostat_template.heating_setpoint = pd.Series([0, 1, np.nan], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='heating_setpoint')
    assert_allclose(sp, 0.9)

    thermostat_template.heating_setpoint = pd.Series([np.nan, np.nan, np.nan], index=index)
    sp = thermostat_template.get_core_heating_day_baseline_setpoint(core_day_set, source='heating_setpoint')
    assert pd.isnull(sp)

def test_get_core_cooling_day_baseline_setpoint_null_cooling_setpoint(thermostat_template):

    index = pd.date_range(start=datetime(2011, 1, 1), periods=3, freq='D')
    thermostat_template.cooling_setpoint = pd.Series([0, 1, 2], index=index)
    thermostat_template.equipment_type = 1

    season_selector = pd.Series([True, True, True], index=index)
    core_day_set = CoreDaySet("FAKE", season_selector, season_selector, None, None)

    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='temperature_in')
    assert pd.isnull(sp)

    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='cooling_setpoint')
    assert_allclose(sp, 0.2)

    thermostat_template.cooling_setpoint = pd.Series([np.nan, 1, 2], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='cooling_setpoint')
    assert_allclose(sp, 1.1)

    thermostat_template.cooling_setpoint = pd.Series([0, 1, np.nan], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='cooling_setpoint')
    assert_allclose(sp, 0.1)

    thermostat_template.cooling_setpoint = pd.Series([np.nan, np.nan, np.nan], index=index)
    sp = thermostat_template.get_core_cooling_day_baseline_setpoint(core_day_set, source='cooling_setpoint')
    assert pd.isnull(sp)



from thermostat.importers import from_csv
from thermostat.util.testing import get_data_path

import numpy as np
import pandas as pd

from datetime import datetime

import pytest

from .fixtures.thermostats import thermostat_type_1
from .fixtures.thermostats import thermostat_type_2
from .fixtures.thermostats import thermostat_type_3
from .fixtures.thermostats import thermostat_type_4
from .fixtures.thermostats import thermostat_type_5
from .fixtures.thermostats import core_heating_day_set_type_1_entire
from .fixtures.thermostats import core_heating_day_set_type_2
from .fixtures.thermostats import core_heating_day_set_type_3
from .fixtures.thermostats import core_heating_day_set_type_4
from .fixtures.thermostats import core_cooling_day_set_type_1_entire
from .fixtures.thermostats import core_cooling_day_set_type_2
from .fixtures.thermostats import core_cooling_day_set_type_3
from .fixtures.thermostats import core_cooling_day_set_type_5
from .fixtures.thermostats import metrics_type_1_data
from .fixtures.thermostats import thermostat_zero_days
from .fixtures.thermostats import thermostats_multiple_same_key

from numpy.testing import assert_allclose
from numpy import isnan


def test_zero_days_warning(thermostat_zero_days):
    output = thermostat_zero_days.calculate_epa_field_savings_metrics()
    assert isnan(output[0]['daily_mean_core_cooling_runtime'])


def test_multiple_same_key(thermostats_multiple_same_key):
    metrics = []
    for thermostat in thermostats_multiple_same_key:
        outputs = thermostat.calculate_epa_field_savings_metrics()
        metrics.extend(outputs)
    assert len(metrics) == 4

    for key in metrics[0]:
        assert(metrics[0][key] == metrics[2][key])

    for key in metrics[1]:
        # nan can never be equal, so skip nan
        if not (isinstance(metrics[1][key], float) and np.isnan(metrics[1][key])):
            assert(metrics[1][key] == metrics[3][key])


def test_interpolate_empty(thermostat_type_1):
    s1 = pd.Series([])
    s1_intp = thermostat_type_1._interpolate(s1)
    np.testing.assert_allclose(s1_intp, [])


def test_interpolate_full(thermostat_type_1):
    s2 = pd.Series([1])
    s2_intp = thermostat_type_1._interpolate(s2)
    np.testing.assert_allclose(s2_intp, [1])
    np.testing.assert_allclose(s2_intp, [1])


def test_interpolate_forward(thermostat_type_1):
    s3 = pd.Series([1, np.nan])
    s3_intp = thermostat_type_1._interpolate(s3)
    np.testing.assert_allclose(s3_intp, [1, 1])


def test_interpolate_backward(thermostat_type_1):
    s4 = pd.Series([np.nan, 1])
    s4_intp = thermostat_type_1._interpolate(s4)
    np.testing.assert_allclose(s4_intp, [1, 1])


def test_interpolate_one_missing(thermostat_type_1):
    s5 = pd.Series([8, 3, np.nan, 1, 7])
    s5_intp = thermostat_type_1._interpolate(s5)
    np.testing.assert_allclose(s5_intp, [8,3,2,1,7])


def test_interpolate_two_missing(thermostat_type_1):
    s6 = pd.Series([8, np.nan, np.nan, 1, 7])
    s6_intp = thermostat_type_1._interpolate(s6)
    np.testing.assert_allclose(s6_intp, [8,5.666,3.333,1,7], rtol=1e-3)


def test_interpolate_three_missing(thermostat_type_1):
    s = pd.Series([8, np.nan, np.nan, np.nan, 1, 7])
    s_intp = thermostat_type_1._interpolate(s)
    np.testing.assert_allclose(s_intp, [8,6.25,np.nan,2.75,1,7])


def test_thermostat_type_1_get_core_heating_days(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_heating_days(
            method="year_mid_to_mid")
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_1_get_core_cooling_days(thermostat_type_1):
    core_cooling_day_sets = thermostat_type_1.get_core_cooling_days(
            method="year_end_to_end")
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_2_get_core_heating_days(thermostat_type_2):
    core_heating_day_sets = thermostat_type_2.get_core_heating_days(
            method="year_mid_to_mid")
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_2_get_core_cooling_days(thermostat_type_2):
    core_cooling_day_sets = thermostat_type_2.get_core_cooling_days(
            method="year_end_to_end")
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_3_get_core_heating_days(thermostat_type_3):
    core_heating_day_sets = thermostat_type_3.get_core_heating_days(
            method="year_mid_to_mid")
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_3_get_core_cooling_days(thermostat_type_3):
    core_cooling_day_sets = thermostat_type_3.get_core_cooling_days(
            method="year_end_to_end")
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_4_get_core_heating_days(thermostat_type_4):
    core_heating_day_sets = thermostat_type_4.get_core_heating_days(
            method="year_mid_to_mid")
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_4_get_core_cooling_days(thermostat_type_4):
    with pytest.raises(ValueError):
        core_cooling_day_sets = thermostat_type_4.get_core_cooling_days(
                method="year_end_to_end")


def test_thermostat_type_5_get_core_heating_days(thermostat_type_5):
    with pytest.raises(ValueError):
        core_heating_day_sets = thermostat_type_5.get_core_heating_days(
                method="year_mid_to_mid")


def test_thermostat_type_5_get_core_cooling_days(thermostat_type_5):
    core_cooling_day_sets = thermostat_type_5.get_core_cooling_days(
            method="year_end_to_end")
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_1_get_core_heating_days_with_params(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_heating_days(
            min_minutes_heating=0, max_minutes_cooling=0,
            method="year_mid_to_mid")
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_1_get_core_cooling_days_with_params(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_cooling_days(
            min_minutes_cooling=0, max_minutes_heating=0,
            method="year_end_to_end")
    assert len(core_heating_day_sets) == 4


def test_thermostat_core_heating_day_set_attributes(core_heating_day_set_type_1_entire):

    assert isinstance(core_heating_day_set_type_1_entire.name, str)
    assert isinstance(core_heating_day_set_type_1_entire.daily, pd.Series)
    assert isinstance(core_heating_day_set_type_1_entire.hourly, pd.Series)
    assert core_heating_day_set_type_1_entire.daily.shape == (1461,)
    assert core_heating_day_set_type_1_entire.hourly.shape == (35064,)
    assert (
        isinstance(core_heating_day_set_type_1_entire.start_date, datetime)
        or isinstance(core_heating_day_set_type_1_entire.start_date, np.datetime64)
    )
    assert (
        isinstance(core_heating_day_set_type_1_entire.end_date, datetime)
        or isinstance(core_heating_day_set_type_1_entire.end_date, np.datetime64)
    )


def test_thermostat_core_cooling_day_set_attributes(core_cooling_day_set_type_1_entire):

    assert isinstance(core_cooling_day_set_type_1_entire.name, str)
    assert isinstance(core_cooling_day_set_type_1_entire.daily, pd.Series)
    assert isinstance(core_cooling_day_set_type_1_entire.hourly, pd.Series)
    assert core_cooling_day_set_type_1_entire.daily.shape == (1461,)
    assert core_cooling_day_set_type_1_entire.hourly.shape == (35064,)
    assert (
        isinstance(core_cooling_day_set_type_1_entire.start_date, datetime)
        or isinstance(core_cooling_day_set_type_1_entire.start_date, np.datetime64)
    )
    assert (
        isinstance(core_cooling_day_set_type_1_entire.end_date, datetime)
        or isinstance(core_cooling_day_set_type_1_entire.end_date, np.datetime64)
    )


def test_thermostat_type_1_total_heating_runtime(thermostat_type_1,
        core_heating_day_set_type_1_entire, metrics_type_1_data):

    total_runtime = thermostat_type_1.total_heating_runtime(core_heating_day_set_type_1_entire)
    assert_allclose(total_runtime, metrics_type_1_data[1]["total_core_heating_runtime"], rtol=1e-3)


def test_thermostat_type_1_total_emergency_heating_runtime(thermostat_type_1,
        core_heating_day_set_type_1_entire, metrics_type_1_data):

    total_runtime = thermostat_type_1.total_emergency_heating_runtime(core_heating_day_set_type_1_entire)
    assert_allclose(total_runtime, metrics_type_1_data[1]["total_emergency_heating_core_day_runtime"], rtol=1e-3)


def test_thermostat_type_1_total_auxiliary_heating_runtime(thermostat_type_1,
        core_heating_day_set_type_1_entire, metrics_type_1_data):

    total_runtime = thermostat_type_1.total_auxiliary_heating_runtime(core_heating_day_set_type_1_entire)
    assert_allclose(total_runtime, metrics_type_1_data[1]["total_auxiliary_heating_core_day_runtime"], rtol=1e-3)


def test_thermostat_type_1_total_cooling_runtime(thermostat_type_1,
        core_cooling_day_set_type_1_entire, metrics_type_1_data):

    total_runtime = thermostat_type_1.total_cooling_runtime(core_cooling_day_set_type_1_entire)
    assert_allclose(total_runtime, metrics_type_1_data[0]["total_core_cooling_runtime"], rtol=1e-3)


def test_thermostat_type_1_get_resistance_heat_utilization_bins_rhu1(thermostat_type_1,
        core_heating_day_set_type_1_entire, metrics_type_1_data):

    start = 0
    stop = 60
    step = 5
    temperature_bins = list(t for t in range(start, stop+step, step))
    rhu_runtime = thermostat_type_1.get_resistance_heat_utilization_runtime(
            core_heating_day_set_type_1_entire)
    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
            rhu_runtime,
            temperature_bins,
            core_heating_day_set_type_1_entire)

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_type_1._format_rhu('rhu1', item.Index.left, item.Index.right, duty_cycle=None)
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_type_1_data[1][bin_name], rtol=1e-3)


def test_thermostat_type_1_get_resistance_heat_utilization_bins_rhu2(thermostat_type_1,
        core_heating_day_set_type_1_entire, metrics_type_1_data):

    start = 0
    stop = 60
    step = 5
    VAR_MIN_RHU_RUNTIME = 30 * 60  # Unit is in minutes (30 hours * 60 minutes)
    temperature_bins = list(t for t in range(start, stop+step, step))
    rhu_runtime = thermostat_type_1.get_resistance_heat_utilization_runtime(
            core_heating_day_set_type_1_entire)
    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
            rhu_runtime,
            temperature_bins,
            core_heating_day_set_type_1_entire,
            VAR_MIN_RHU_RUNTIME)

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_type_1._format_rhu('rhu2', item.Index.left, item.Index.right, duty_cycle=None)
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_type_1_data[1][bin_name], rtol=1e-3)


def test_thermostat_types_2_get_resistance_heat_utilization_runtime_rhu(
        core_heating_day_set_type_1_entire, thermostat_type_2):

    with pytest.raises(ValueError):
        thermostat_type_2.get_resistance_heat_utilization_runtime(
            core_heating_day_set_type_1_entire)


@pytest.fixture(params=range(2))
def core_days(request, thermostat_type_1, core_heating_day_set_type_1_entire,
        core_cooling_day_set_type_1_entire):

    tests = [
        (thermostat_type_1, core_cooling_day_set_type_1_entire, 0, "cooling"),
        (thermostat_type_1, core_heating_day_set_type_1_entire, 1, "heating"),
    ]

    return tests[request.param]


def test_day_counts(core_days, metrics_type_1_data):
    thermostat, core_day_set, i, heating_or_cooling = core_days
    n_both, n_days_insufficient = thermostat.get_ignored_days(core_day_set)
    n_core_days = thermostat.get_core_day_set_n_days(core_day_set)
    n_days_in_inputfile_date_range = thermostat.get_inputfile_date_range(core_day_set)
    assert n_both == metrics_type_1_data[i]["n_days_both_heating_and_cooling"]
    assert n_days_insufficient == metrics_type_1_data[i]["n_days_insufficient_data"]
    assert n_core_days == metrics_type_1_data[i]["n_core_{}_days".format(heating_or_cooling)]
    assert n_days_in_inputfile_date_range == metrics_type_1_data[i]["n_days_in_inputfile_date_range"]



from thermostat.importers import from_csv
from thermostat.util.testing import get_data_path
from thermostat.regression import runtime_regression
from thermostat.core import Thermostat, CoreDaySet

import pandas as pd
import numpy as np
from numpy import nan

import pytest

# will be modified, recreate every time by scoping to function
@pytest.fixture(scope='function')
def thermostat_template():
    thermostat_id = "FAKE"
    equipment_type = 0
    zipcode = "FAKE"
    station = "FAKE"
    temperature_in = pd.Series([])
    temperature_out = pd.Series([])
    cooling_setpoint = pd.Series([])
    heating_setpoint = pd.Series([])
    cool_runtime = pd.Series([])
    heat_runtime = pd.Series([])
    auxiliary_heat_runtime = pd.Series([])
    emergency_heat_runtime = pd.Series([])

    thermostat = Thermostat(
        thermostat_id,
        equipment_type,
        zipcode,
        station,
        temperature_in,
        temperature_out,
        cooling_setpoint,
        heating_setpoint,
        cool_runtime,
        heat_runtime,
        auxiliary_heat_runtime,
        emergency_heat_runtime
    )
    return thermostat

# Note:
# The following fixtures can be quite slow without a prebuilt weather cache
# they the from_csv command fetches weather data. (This happens with builds on
# travis.)
# To speed this up, spoof the weather source.

@pytest.fixture(scope="session", params=["../data/metadata_type_1_single_utc_offset_0.csv"])
def thermostat_type_1_utc(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_type_1_single_utc_offset_bad.csv"])
def thermostat_type_1_utc_bad(request):
    thermostats = from_csv(get_data_path(request.param))

@pytest.fixture(scope="session", params=["../data/metadata_multiple_same_key.csv"])
def thermostats_multiple_same_key(request):
    thermostats = from_csv(get_data_path(request.param))
    return thermostats

@pytest.fixture(scope="session", params=["../data/metadata_type_1_single.csv"])
def thermostat_type_1(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_type_2_single.csv"])
def thermostat_type_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_type_3_single.csv"])
def thermostat_type_3(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_type_4_single.csv"])
def thermostat_type_4(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_type_5_single.csv"])
def thermostat_type_5(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_single_zero_days.csv"])
def thermostat_zero_days(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/metadata_single_emg_aux_constant_on_outlier.csv"])
def thermostat_emg_aux_constant_on_outlier(request):
    thermostats = from_csv(get_data_path(request.param))
    return thermostats

@pytest.fixture(scope="session")
def core_heating_day_set_type_1_mid_to_mid(thermostat_type_1):
    return thermostat_type_1.get_core_heating_days(method="year_mid_to_mid")[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_heating_days(method="entire_dataset")[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_heating_days(method="year_mid_to_mid")[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_heating_days(method="year_mid_to_mid")[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_4(thermostat_type_4):
    return thermostat_type_4.get_core_heating_days(method="year_mid_to_mid")[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_end_to_end(thermostat_type_1):
    return thermostat_type_1.get_core_cooling_days(method="year_end_to_end")[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_cooling_days(method="entire_dataset")[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_empty(thermostat_type_1):
    core_cooling_day_set = thermostat_type_1.get_core_cooling_days(method="entire_dataset")[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(np.tile(False, core_cooling_day_set.daily.shape),
                  index=core_cooling_day_set.daily.index),
        pd.Series(np.tile(False, core_cooling_day_set.hourly.shape),
                  index=core_cooling_day_set.hourly.index),
        core_cooling_day_set.start_date,
        core_cooling_day_set.end_date
    )
    return core_day_set

@pytest.fixture(scope="session")
def core_heating_day_set_type_1_empty(thermostat_type_1):
    core_heating_day_set = thermostat_type_1.get_core_heating_days(method="entire_dataset")[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(np.tile(False, core_heating_day_set.daily.shape),
                  index=core_heating_day_set.daily.index),
        pd.Series(np.tile(False, core_heating_day_set.hourly.shape),
                  index=core_heating_day_set.hourly.index),
        core_heating_day_set.start_date,
        core_heating_day_set.end_date
    )
    return core_day_set

@pytest.fixture(scope="session")
def core_cooling_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_cooling_days(method="year_end_to_end")[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_cooling_days(method="year_end_to_end")[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_5(thermostat_type_5):
    return thermostat_type_5.get_core_cooling_days(method="year_end_to_end")[0]

@pytest.fixture(scope="session")
def metrics_type_1_data():

    # this data comes from a script in scripts/test_data_generation.ipynb
    data = [{
        'sw_version': '1.7.5',
        'ct_identifier': '8465829e-df0d-449e-97bf-96317c24dec3',
        'equipment_type': 1,
        'heating_or_cooling': 'cooling_ALL',
        'zipcode': '62223',
        'station': '725314',
        'climate_zone': 'Mixed-Humid',
        'start_date': '2011-01-01T00:00:00',
        'end_date': '2014-12-31T00:00:00',
        'n_days_in_inputfile_date_range': 1460,
        'n_days_both_heating_and_cooling': 212,
        'n_days_insufficient_data': 13,
        'n_core_cooling_days': 296,
        'baseline_percentile_core_cooling_comfort_temperature': 69.5,
        'regional_average_baseline_cooling_comfort_temperature': 73.0,
        'percent_savings_baseline_percentile': 43.83576234744179,
        'avoided_daily_mean_core_day_runtime_baseline_percentile': 192.71571034192178,
        'avoided_total_core_day_runtime_baseline_percentile': 57043.85026120885,
        'baseline_daily_mean_core_day_runtime_baseline_percentile': 439.6312508824623,
        'baseline_total_core_day_runtime_baseline_percentile': 130130.85026120885,
        '_daily_mean_core_day_demand_baseline_baseline_percentile': 9.976082164677136,
        'percent_savings_baseline_regional': 21.403521446912517,
        'avoided_daily_mean_core_day_runtime_baseline_regional': 67.24044339932902,
        'avoided_total_core_day_runtime_baseline_regional': 19903.17124620139,
        'baseline_daily_mean_core_day_runtime_baseline_regional': 314.15598393986954,
        'baseline_total_core_day_runtime_baseline_regional': 92990.17124620138,
        '_daily_mean_core_day_demand_baseline_baseline_regional': 7.128806020996521,
        'mean_demand': 5.60299049538374,
        'tau': -0.8165326352929079,
        'alpha': 44.06852746653279,
        'mean_sq_err': 379.28224705229485,
        'root_mean_sq_err': 19.47517001343749,
        'cv_root_mean_sq_err': 0.07887381236030344,
        'mean_abs_pct_err': 0.050400459129042466,
        'mean_abs_err': 12.444656609338942,
        'total_core_cooling_runtime': 73087.0,
        'daily_mean_core_cooling_runtime': 246.91554054054055,
        'core_cooling_days_mean_indoor_temperature': 73.95971753003002,
        'core_cooling_days_mean_outdoor_temperature': 79.8426321875,
        'core_mean_indoor_temperature': 73.95971753003002,
        'core_mean_outdoor_temperature': 79.8426321875
        }, {
        'sw_version': '1.7.5',
        'ct_identifier': '8465829e-df0d-449e-97bf-96317c24dec3',
        'equipment_type': 1,
        'heating_or_cooling': 'heating_ALL',
        'zipcode': '62223',
        'station': '725314',
        'climate_zone': 'Mixed-Humid',
        'start_date': '2011-01-01T00:00:00',
        'end_date': '2014-12-31T00:00:00',
        'n_days_in_inputfile_date_range': 1460,
        'n_days_both_heating_and_cooling': 212,
        'n_days_insufficient_data': 13,
        'n_core_heating_days': 895,
        'baseline_percentile_core_heating_comfort_temperature': 69.5,
        'regional_average_baseline_heating_comfort_temperature': 69,
        'percent_savings_baseline_percentile': 10.64692893106402,
        'avoided_daily_mean_core_day_runtime_baseline_percentile': 92.11668592857713,
        'avoided_total_core_day_runtime_baseline_percentile': 82444.43390607653,
        'baseline_daily_mean_core_day_runtime_baseline_percentile': 865.19489821908,
        'baseline_total_core_day_runtime_baseline_percentile': 774349.4339060766,
        '_daily_mean_core_day_demand_baseline_baseline_percentile': 27.292461874951055,
        'percent_savings_baseline_regional': 9.033255812379368,
        'avoided_daily_mean_core_day_runtime_baseline_regional': 76.76886005938185,
        'avoided_total_core_day_runtime_baseline_regional': 68708.12975314676,
        'baseline_daily_mean_core_day_runtime_baseline_regional': 849.8470723498847,
        'baseline_total_core_day_runtime_baseline_regional': 760613.1297531468,
        '_daily_mean_core_day_demand_baseline_baseline_regional': 26.808316680312686,
        'mean_demand': 24.386652855587275,
        'tau': -2.3486605751007135,
        'alpha': 31.700874116202517,
        'mean_sq_err': 6897.07048492154,
        'root_mean_sq_err': 83.04860314852706,
        'cv_root_mean_sq_err': 0.1074258746763381,
        'mean_abs_pct_err': 0.07192178838260584,
        'mean_abs_err': 55.60116758756078,
        'total_core_heating_runtime': 691905.0,
        'daily_mean_core_heating_runtime': 773.0782122905028,
        'core_heating_days_mean_indoor_temperature': 66.6958723285375,
        'core_heating_days_mean_outdoor_temperature': 44.679677289106145,
        'core_mean_indoor_temperature': 66.6958723285375,
        'core_mean_outdoor_temperature': 44.679677289106145,
        'total_auxiliary_heating_core_day_runtime': 144794.0,
        'total_emergency_heating_core_day_runtime': 2104.0,
        'rhu1_aux_duty_cycle': 0.06990582191780823,
        'rhu1_emg_duty_cycle': 0.0010435692541856925,
        'rhu1_compressor_duty_cycle': 0.3380902777777778,
        'rhu1_00F_to_05F': nan,
        'rhu1_05F_to_10F': 0.35810185185185184,
        'rhu1_10F_to_15F': 0.36336805555555557,
        'rhu1_15F_to_20F': 0.37900691389063484,
        'rhu1_20F_to_25F': 0.375673443706005,
        'rhu1_25F_to_30F': 0.3318400322150354,
        'rhu1_30F_to_35F': 0.28432496802364043,
        'rhu1_35F_to_40F': 0.19741898266605878,
        'rhu1_40F_to_45F': 0.15271363589013506,
        'rhu1_45F_to_50F': 0.09249776186213071,
        'rhu1_50F_to_55F': 0.052322643343051506,
        'rhu1_55F_to_60F': 0.028319891645631964,
        'rhu1_less10F': 0.3651785714285714,
        'rhu1_10F_to_20F': 0.3745206434583396,
        'rhu1_20F_to_30F': 0.3489457045517326,
        'rhu1_30F_to_40F': 0.24596842083649878,
        'rhu1_40F_to_50F': 0.12306960016924054,
        'rhu1_50F_to_60F': 0.0430327868852459,
        'rhu1_00F_to_05F_aux_duty_cycle': nan,
        'rhu1_05F_to_10F_aux_duty_cycle': 0.35810185185185184,
        'rhu1_10F_to_15F_aux_duty_cycle': 0.36336805555555557,
        'rhu1_15F_to_20F_aux_duty_cycle': 0.3760763888888889,
        'rhu1_20F_to_25F_aux_duty_cycle': 0.35656906906906904,
        'rhu1_25F_to_30F_aux_duty_cycle': 0.3082627118644068,
        'rhu1_30F_to_35F_aux_duty_cycle': 0.23578539823008848,
        'rhu1_35F_to_40F_aux_duty_cycle': 0.14188453159041395,
        'rhu1_40F_to_45F_aux_duty_cycle': 0.08783360566448802,
        'rhu1_45F_to_50F_aux_duty_cycle': 0.043188657407407405,
        'rhu1_50F_to_55F_aux_duty_cycle': 0.01672854785478548,
        'rhu1_55F_to_60F_aux_duty_cycle': 0.005432581018518519,
        'rhu1_less10F_aux_duty_cycle': 0.3651785714285714,
        'rhu1_10F_to_20F_aux_duty_cycle': 0.3724454365079365,
        'rhu1_20F_to_30F_aux_duty_cycle': 0.32688078703703705,
        'rhu1_30F_to_40F_aux_duty_cycle': 0.19123708010335919,
        'rhu1_40F_to_50F_aux_duty_cycle': 0.06370120120120121,
        'rhu1_50F_to_60F_aux_duty_cycle': 0.011223914269599549,
        'rhu1_00F_to_05F_emg_duty_cycle': nan,
        'rhu1_05F_to_10F_emg_duty_cycle': 0.0,
        'rhu1_10F_to_15F_emg_duty_cycle': 0.0,
        'rhu1_15F_to_20F_emg_duty_cycle': 0.0007986111111111112,
        'rhu1_20F_to_25F_emg_duty_cycle': 0.002027027027027027,
        'rhu1_25F_to_30F_emg_duty_cycle': 0.00211864406779661,
        'rhu1_30F_to_35F_emg_duty_cycle': 0.0019174041297935103,
        'rhu1_35F_to_40F_emg_duty_cycle': 0.002573529411764706,
        'rhu1_40F_to_45F_emg_duty_cycle': 0.001994825708061002,
        'rhu1_45F_to_50F_emg_duty_cycle': 0.0016550925925925926,
        'rhu1_50F_to_55F_emg_duty_cycle': 0.0017808030803080307,
        'rhu1_55F_to_60F_emg_duty_cycle': 0.001222511574074074,
        'rhu1_less10F_emg_duty_cycle': 0.0,
        'rhu1_10F_to_20F_emg_duty_cycle': 0.000570436507936508,
        'rhu1_20F_to_30F_emg_duty_cycle': 0.0020833333333333333,
        'rhu1_30F_to_40F_emg_duty_cycle': 0.0022286821705426356,
        'rhu1_40F_to_50F_emg_duty_cycle': 0.0018111861861861863,
        'rhu1_50F_to_60F_emg_duty_cycle': 0.001508742244782854,
        'rhu1_00F_to_05F_compressor_duty_cycle': nan,
        'rhu1_05F_to_10F_compressor_duty_cycle': 1.0,
        'rhu1_10F_to_15F_compressor_duty_cycle': 1.0,
        'rhu1_15F_to_20F_compressor_duty_cycle': 0.9935763888888889,
        'rhu1_20F_to_25F_compressor_duty_cycle': 0.952515015015015,
        'rhu1_25F_to_30F_compressor_duty_cycle': 0.9332156308851224,
        'rhu1_30F_to_35F_compressor_duty_cycle': 0.8341076696165192,
        'rhu1_35F_to_40F_compressor_duty_cycle': 0.7291598583877995,
        'rhu1_40F_to_45F_compressor_duty_cycle': 0.5862200435729847,
        'rhu1_45F_to_50F_compressor_duty_cycle': 0.4831539351851852,
        'rhu1_50F_to_55F_compressor_duty_cycle': 0.3519733223322332,
        'rhu1_55F_to_60F_compressor_duty_cycle': 0.2337745949074074,
        'rhu1_less10F_compressor_duty_cycle': 1.0,
        'rhu1_10F_to_20F_compressor_duty_cycle': 0.9954117063492064,
        'rhu1_20F_to_30F_compressor_duty_cycle': 0.9406539351851851,
        'rhu1_30F_to_40F_compressor_duty_cycle': 0.7843184754521964,
        'rhu1_40F_to_50F_compressor_duty_cycle': 0.5305086336336337,
        'rhu1_50F_to_60F_compressor_duty_cycle': 0.29437394247038917,
        'rhu2_aux_duty_cycle': 0.06990582191780823,
        'rhu2_emg_duty_cycle': 0.0010435692541856925,
        'rhu2_compressor_duty_cycle': 0.3380902777777778,
        'rhu2_00F_to_05F': nan,
        'rhu2_05F_to_10F': 0.35810185185185184,
        'rhu2_10F_to_15F': 0.36336805555555557,
        'rhu2_15F_to_20F': 0.37900691389063484,
        'rhu2_20F_to_25F': 0.375673443706005,
        'rhu2_25F_to_30F': 0.3318400322150354,
        'rhu2_30F_to_35F': 0.28432496802364043,
        'rhu2_35F_to_40F': 0.19741898266605878,
        'rhu2_40F_to_45F': 0.15271363589013506,
        'rhu2_45F_to_50F': 0.09249776186213071,
        'rhu2_50F_to_55F': 0.052322643343051506,
        'rhu2_55F_to_60F': 0.028319891645631964,
        'rhu2_less10F': 0.3651785714285714,
        'rhu2_10F_to_20F': 0.3745206434583396,
        'rhu2_20F_to_30F': 0.3489457045517326,
        'rhu2_30F_to_40F': 0.24596842083649878,
        'rhu2_40F_to_50F': 0.12306960016924054,
        'rhu2_50F_to_60F': 0.0430327868852459,
        'rhu2_00F_to_05F_aux_duty_cycle': nan,
        'rhu2_05F_to_10F_aux_duty_cycle': 0.35810185185185184,
        'rhu2_10F_to_15F_aux_duty_cycle': 0.36336805555555557,
        'rhu2_15F_to_20F_aux_duty_cycle': 0.3760763888888889,
        'rhu2_20F_to_25F_aux_duty_cycle': 0.35656906906906904,
        'rhu2_25F_to_30F_aux_duty_cycle': 0.3082627118644068,
        'rhu2_30F_to_35F_aux_duty_cycle': 0.23578539823008848,
        'rhu2_35F_to_40F_aux_duty_cycle': 0.14188453159041395,
        'rhu2_40F_to_45F_aux_duty_cycle': 0.08783360566448802,
        'rhu2_45F_to_50F_aux_duty_cycle': 0.043188657407407405,
        'rhu2_50F_to_55F_aux_duty_cycle': 0.01672854785478548,
        'rhu2_55F_to_60F_aux_duty_cycle': 0.005432581018518519,
        'rhu2_less10F_aux_duty_cycle': 0.3651785714285714,
        'rhu2_10F_to_20F_aux_duty_cycle': 0.3724454365079365,
        'rhu2_20F_to_30F_aux_duty_cycle': 0.32688078703703705,
        'rhu2_30F_to_40F_aux_duty_cycle': 0.19123708010335919,
        'rhu2_40F_to_50F_aux_duty_cycle': 0.06370120120120121,
        'rhu2_50F_to_60F_aux_duty_cycle': 0.011223914269599549,
        'rhu2_00F_to_05F_emg_duty_cycle': nan,
        'rhu2_05F_to_10F_emg_duty_cycle': 0.0,
        'rhu2_10F_to_15F_emg_duty_cycle': 0.0,
        'rhu2_15F_to_20F_emg_duty_cycle': 0.0007986111111111112,
        'rhu2_20F_to_25F_emg_duty_cycle': 0.002027027027027027,
        'rhu2_25F_to_30F_emg_duty_cycle': 0.00211864406779661,
        'rhu2_30F_to_35F_emg_duty_cycle': 0.0019174041297935103,
        'rhu2_35F_to_40F_emg_duty_cycle': 0.002573529411764706,
        'rhu2_40F_to_45F_emg_duty_cycle': 0.001994825708061002,
        'rhu2_45F_to_50F_emg_duty_cycle': 0.0016550925925925926,
        'rhu2_50F_to_55F_emg_duty_cycle': 0.0017808030803080307,
        'rhu2_55F_to_60F_emg_duty_cycle': 0.001222511574074074,
        'rhu2_less10F_emg_duty_cycle': 0.0,
        'rhu2_10F_to_20F_emg_duty_cycle': 0.000570436507936508,
        'rhu2_20F_to_30F_emg_duty_cycle': 0.0020833333333333333,
        'rhu2_30F_to_40F_emg_duty_cycle': 0.0022286821705426356,
        'rhu2_40F_to_50F_emg_duty_cycle': 0.0018111861861861863,
        'rhu2_50F_to_60F_emg_duty_cycle': 0.001508742244782854,
        'rhu2_00F_to_05F_compressor_duty_cycle': nan,
        'rhu2_05F_to_10F_compressor_duty_cycle': 1.0,
        'rhu2_10F_to_15F_compressor_duty_cycle': 1.0,
        'rhu2_15F_to_20F_compressor_duty_cycle': 0.9935763888888889,
        'rhu2_20F_to_25F_compressor_duty_cycle': 0.952515015015015,
        'rhu2_25F_to_30F_compressor_duty_cycle': 0.9332156308851224,
        'rhu2_30F_to_35F_compressor_duty_cycle': 0.8341076696165192,
        'rhu2_35F_to_40F_compressor_duty_cycle': 0.7291598583877995,
        'rhu2_40F_to_45F_compressor_duty_cycle': 0.5862200435729847,
        'rhu2_45F_to_50F_compressor_duty_cycle': 0.4831539351851852,
        'rhu2_50F_to_55F_compressor_duty_cycle': 0.3519733223322332,
        'rhu2_55F_to_60F_compressor_duty_cycle': 0.2337745949074074,
        'rhu2_less10F_compressor_duty_cycle': 1.0,
        'rhu2_10F_to_20F_compressor_duty_cycle': 0.9954117063492064,
        'rhu2_20F_to_30F_compressor_duty_cycle': 0.9406539351851851,
        'rhu2_30F_to_40F_compressor_duty_cycle': 0.7843184754521964,
        'rhu2_40F_to_50F_compressor_duty_cycle': 0.5305086336336337,
        'rhu2_50F_to_60F_compressor_duty_cycle': 0.29437394247038917}]
    return data



from multiprocessing import Pool


def _calc_epa_func(thermostat):
    """ Takes an individual thermostat and runs the
    calculate_epa_field_savings_metrics method. This method is necessary for
    the multiprocessing pool as map / imap need a function to run on.

    Parameters
    ----------
    thermostat : thermostat

    Returns
    -------
    results : results from running calculate_epa_field_savings_metrics
    """
    results = thermostat.calculate_epa_field_savings_metrics()
    return results


def multiple_thermostat_calculate_epa_field_savings_metrics(thermostats):
    """ Takes a list of thermostats and uses Python's Multiprocessing module to
    run as many processes in parallel as the system will allow.

    Parameters
    ----------
    thermostats : thermostats iterator
        A list of the thermostats run the calculate_epa_field_savings_metrics
        upon.

    Returns
    -------
    metrics : list
        Returns a list of the metrics calculated for the thermostats
    """
    # Convert the thermostats iterator to a list
    thermostats_list = list(thermostats)

    pool = Pool()
    results = pool.imap(_calc_epa_func, thermostats_list)
    pool.close()
    pool.join()

    metrics_dict = {}
    for output in results:
        thermostat_id = output[0]['ct_identifier']
        metrics_dict[thermostat_id] = []
        for individual_output in output:
            metrics_dict[thermostat_id].append(individual_output)

    # Get the order of the thermostats from the original input so the output
    # matches the order that was sent in
    thermostat_ids = \
        [thermostat.thermostat_id for thermostat in thermostats_list]
    metrics = []
    for thermostat_id in thermostat_ids:
        try:
            for metric in metrics_dict[thermostat_id]:
                metrics.append(metric)
            # Prevent duplicate thermostat IDs from being double-counted
            metrics_dict.pop(thermostat_id, None)
        # Trap for missing keys
        except KeyError:
            pass

    return metrics
